import tkinter as tk
from tkinter import filedialog, ttk, messagebox
import os
from msiAlign.downcore_profile import calc_depth_profile_combined
from msiAlign.metadata_crawler import crawl_metadata
from msiAlign.objects import MsiImage, TeachableImage
from msiAlign.shinyApp import run_app


def how_to_use():
    """open a webpage to show how to use the software"""
    try:
        import webbrowser
        webbrowser.open("https://github.com/weimin-liu/msiAlign/blob/main/README.md")
    except Exception as e:
        window = tk.Toplevel()
        window.title("How to use")
        text = tk.Text(window)
        text.insert(tk.END, "Please visit https://github.com/weimin-liu/msiAlign/blob/main/README.md")


class MenuBar:

    def __init__(self, app):
        self.app = app


        # create a menubar
        self.menubar = tk.Menu(self.app)
        self.app.config(menu=self.menubar)

        # Add file menu
        self.file_menu = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="File", menu=self.file_menu)
        self.file_menu.add_command(label="Add images", command=self.add_images, accelerator="Ctrl+O")
        self.app.bind("<Control-o>", self.add_images)
        self.file_menu.add_separator()
        self.file_menu.add_command(label="Save workspace", command=self.app.save, accelerator="Ctrl+S")
        self.app.bind("<Control-s>", self.app.save)
        self.file_menu.add_command(label="Save layout", command=self.app.save_layout)
        self.file_menu.add_command(label="Load workspace", command=self.app.load, accelerator="Ctrl+L")
        self.app.bind("<Control-l>", self.app.load)
        self.file_menu.add_separator()
        # Add 'Open' to the file menu

        # Add 'crawl metadata' to the file menu
        self.file_menu.add_command(label="Crawl metadata", command=crawl_metadata)
        # Add 'Add metadata' to the file menu
        self.file_menu.add_command(label="Attach database", command=self.app.add_metadata)
        # Add 'Exit' to the file menu
        self.file_menu.add_separator()
        self.file_menu.add_command(label="Quit", command=self.quit, accelerator="Ctrl+Q")
        self.app.bind("<Control-q>", self.quit)

        # Add 'View' menu
        self.view_menu = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="View", menu=self.view_menu)
        # show all tp labels next to the teaching points
        self.view_menu.add_command(label="Show TP Labels", command=self.app.show_tp_labels)
        # hide all tp labels next to the teaching points
        self.view_menu.add_command(label="Hide TP Labels", command=self.app.hide_tp_labels)
        self.view_menu.add_separator()
        # add 'Hide Teaching Points View' to the view menu
        self.view_menu.add_command(label="Toggle TP View", command=self.tg_tp_view)
        # update the teaching points view
        self.view_menu.add_command(label="Update TP View", command=self.update_tp_view)


        # Add 'Calc' menu
        self.calc_menu = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="Calc", menu=self.calc_menu)
        # Add 'Cm/Px' to the calc menu
        self.calc_menu.add_command(label="cm/px", command=self.app.calculation_handler.calc_cm_per_px)
        self.calc_menu.add_separator()
        # calculate the MSI machine coordinate
        # calculate the transformation matrix
        # convert the machine coordinate to real world coordinate
        self.calc_menu.add_command(label="Prep MSI", command=self.app.calculation_handler.calc_msi)

        self.calc_menu.add_command(label="Prep XRF", command=lambda: self.pair_tps_ui(xrf=True))

        self.calc_menu.add_separator()
        self.calc_menu.add_command(label="Downcore Profile", command=calc_depth_profile_combined)

        self.calc_menu.add_command(label="Downcore Profile (new)", command=run_app)

        # Add a 'Dev' menu
        self.dev_menu = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="Dev", menu=self.dev_menu)
        self.menubar.entryconfig("Dev", state="disabled")
        # bind Ctrl+p to enable the dev menu
        self.app.bind("<Control-p>", self.enable_dev_menu)

        # Add labels for all the teaching points
        self.dev_menu.add_command(label="Reset TP", command=self.app.dev_ops_handler.reset_tp)
        self.dev_menu.add_command(label="Reset All", command=self.app.dev_ops_handler.reset)
        self.dev_menu.add_command(label="Set TP Size", command=self.app.dev_ops_handler.set_tp_size)
        # move all teaching points to the top of the canvas
        self.dev_menu.add_command(label="Export TPs", command=self.app.dev_ops_handler.export_tps)
        # calculate the depth for all teaching points
        self.dev_menu.add_command(label="Calc Depth for All TPs", command=self.app.dev_ops_handler.calc_depth_for_all_tps)
        # lock all the images
        self.dev_menu.add_command(label="Lock All Images", command=self.app.dev_ops_handler.lock_all)


        # Add 'Help' menu
        self.help_menu = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="Help", menu=self.help_menu)
        # Add 'About' to the help menu
        self.help_menu.add_command(label="How to use", command=how_to_use)
        # Add an 'Issue' to the help menu
        self.help_menu.add_command(label="Report an issue", command=report_issue)


    def enable_dev_menu(self, event=None):
        """Enable the dev menu"""
        self.menubar.entryconfig("Dev", state="normal")
        # add a message pop-up to show the dev menu is enabled
        messagebox.showinfo("Dev Menu", "Developer menu is enabled")
        # unbind the Ctrl+p to prevent the dev menu from being enabled again
        self.app.unbind("<Control-p>")
        return "break"


    def add_images(self, event=None):
        """Load the images from the file paths"""
        file_paths = filedialog.askopenfilenames(title="Select image files")

        for file_path in file_paths:
            for k, v in self.app.items.items():
                try:
                    if k == os.path.basename(file_path):
                        messagebox.showerror("Error", f"{file_path} has already been loaded")
                        return
                except AttributeError:
                    pass
            # let the user choose if input image is a reference image or a MSI image
            image_type = messagebox.askyesno(
                "Image Type", "Is this a MSI image?"
            )  # if the user cancels, the image is a msi image
            if image_type:
                loaded_image = MsiImage.from_path(file_path)
            else:
                loaded_image = TeachableImage.from_path(file_path)
            loaded_image.create_im_on_canvas(self.app)
            self.app.items[loaded_image.tag] = loaded_image

    def quit(self, event=None):
        self.app.quit()

    def generate_tree_view(self):
        """generate a tree view using the teaching points, with the parent node being the image tag"""
        # clear the tree view
        self.app.tree.delete(*self.app.tree.get_children())
        # get the teaching points
        for k, v in self.app.items.items():
            try:
                if v.teaching_points is not None:
                    # add the image tag to the tree view
                    parent = self.app.tree.insert("", "end", text=k, values=("image", "", "", ""))
                    # add the teaching points to the tree view
                    for i, tp in v.teaching_points.items():
                        self.app.tree.insert(parent, "end", text=i, values=("", tp[0], tp[1], tp[2]))
            except AttributeError:
                pass

    def update_tp_view(self):
        # if the teaching points view does not exist, create one
        if not hasattr(self.app, 'tree'):
            # create a treeview to display teaching points
            self.app.tree_frame = tk.Frame(self.app)
            self.app.tree_frame.pack(side=tk.RIGHT, fill=tk.Y)
            self.app.tree = ttk.Treeview(self.app.tree_frame,
                                         columns=('label', 'x', 'y', 'd'),
                                         selectmode='browse')
            self.app.tree.heading('#0', text='img')
            self.app.tree.heading('label', text='label')
            self.app.tree.heading('x', text='x')
            self.app.tree.heading('y', text='y')
            self.app.tree.heading('d', text='d')
            self.app.tree.column('#0', width=100)
            self.app.tree.column('label', width=50)
            self.app.tree.column('x', width=50)
            self.app.tree.column('y', width=50)
            self.app.tree.column('d', width=50)
            self.app.tree.pack(side=tk.LEFT, fill=tk.Y)
            self.app.tree_visible = True
        # update the teaching points view
        self.generate_tree_view()

    def tg_tp_view(self):
        """Toggle the visibility of the teaching points view"""
        if self.app.tree_visible:
            self.app.tree_frame.pack_forget()
            self.app.tree_visible = False
        else:
            self.app.tree_frame.pack(side=tk.RIGHT, fill=tk.Y)
            self.app.tree_visible = True

    def pair_tps_ui(self, msi=False,xrf=False):
        """pair the teaching points from xray images and msi images"""
        # create a pop-up text window to input the pair of teaching points
        popup = tk.Toplevel()
        popup.title("Pair Teaching Points")
        popup.geometry("300x200")
        # create a text input box to input the pair of teaching points that are strecthable by the user
        text = tk.Text(popup, height=10, width=30)
        text.grid(row=0, column=0, sticky="nsew")
        popup.grid_columnconfigure(0, weight=1)
        popup.grid_rowconfigure(0, weight=1)
        auto_label_button = tk.Button(popup, text="Label all TPs",
                                      command=lambda: self.app.add_tp_labels())
        auto_label_button.grid(row=1, column=0, sticky="nsew")

        # create a button to submit the pair of teaching points
        submit_button = tk.Button(popup, text="Submit",
                                  command=lambda: self.app.calculation_handler.pair_tps(text.get("1.0", "end-1c"),
                                                                    msi=msi,
                                                                    xrf=xrf))
        submit_button.grid(row=2, column=0, sticky="nsew")
        # create a button to fill in the pair of teaching points
        fill_button = tk.Button(popup, text="Fill", command=lambda: text.insert(tk.END, self.app.fill_tps_str()))
        fill_button.grid(row=3, column=0, sticky="nsew")




def report_issue():
    try:
        import webbrowser
        webbrowser.open("https://github.com/weimin-liu/msiAlign/issues")
    except Exception as e:
        messagebox.showinfo("Report an issue",
                            "Please report the issue at https://github.com/weimin-liu/msiAlign/issues")




""" crawler for metadata """
import os
import shutil
# create a namedtuple for the metadata
from collections import namedtuple
from tkinter import filedialog, simpledialog, messagebox
from msiAlign.func import get_image_file_from_mis, get_px_rect_from_mis, get_msi_rect_from_imaginginfo

Metadata = namedtuple("Metadata", ["spec_file_name",
                                   "msi_img_file_path",
                                   "msi_img_file_name",
                                   "px_rect",
                                   "msi_rect",
                                   "spot_name",
                                   "TIC",
                                   "maxpeak",
                                   "rt"])


class MetadataCrawler:
    """Crawl the metadata directory and return the metadata files"""

    def __init__(self, metadata_dir):
        self.metadata_dir = metadata_dir
        self.metadata = {}
        # make sure the subdirectories are something ending with '.i'
        self.idir = [subdir for subdir in os.listdir(metadata_dir) if subdir.endswith(".i")]
        assert len(self.idir) > 0, "No .i subdirectories found"

    def crawl(self):
        # craw the metadata under .i subdirectories
        # get all the .mis files under the .i subdirectories
        for subdir in self.idir:
            for root, dirs, files in os.walk(os.path.join(self.metadata_dir, subdir)):
                for file in files:
                    if file.endswith(".mis"):
                        mis_file = os.path.join(root, file)
                        # infer the spec file name from the mis file
                        spec_file_name = mis_file.replace(".mis", ".d")
                        if os.path.exists(spec_file_name):
                            try:
                                px_rect = get_px_rect_from_mis(mis_file)
                                xml_file = os.path.join(root, spec_file_name, "ImagingInfo.xml")
                                msi_rect, spot_name, tic, maxpeak, rt = get_msi_rect_from_imaginginfo(xml_file, return_spot_name=True)
                                im_name = get_image_file_from_mis(mis_file)
                                im_file_path = os.path.join(root, im_name)
                                assert im_name not in self.metadata.keys(), 'duplicate entries found'
                                self.metadata[os.path.basename(spec_file_name)] = Metadata(
                                    spec_file_name=os.path.basename(spec_file_name),
                                    msi_img_file_path=im_file_path,
                                    msi_img_file_name=im_name,
                                    px_rect=px_rect,
                                    msi_rect=msi_rect,
                                    spot_name=spot_name,
                                    TIC=tic,
                                    maxpeak=maxpeak,
                                    rt=rt)
                            except ValueError:
                                continue

    def to_sqlite(self, db_path):
        # create a sqlite database if it does not exist
        if not os.path.exists(db_path):
            open(db_path, 'w').close()
        else:
            os.remove(db_path)
            open(db_path, 'w').close()
        # write the metadata to the sqlite database
        import sqlite3
        conn = sqlite3.connect(db_path)
        c = conn.cursor()
        c.execute(
            'CREATE TABLE IF NOT EXISTS metadata (spec_id INTEGER PRIMARY KEY, spec_file_name TEXT, msi_img_file_path TEXT, '
            'msi_img_file_name TEXT, px_rect TEXT, msi_rect TEXT, spot_name TEXT, TIC TEXT, maxpeak TEXT, rt TEXT)')
        for k, v in self.metadata.items():
            c.execute(
                'INSERT INTO metadata (spec_file_name, msi_img_file_path, msi_img_file_name, px_rect, msi_rect, spot_name, TIC, maxpeak, rt) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)',
                (v.spec_file_name, v.msi_img_file_path,
                 v.msi_img_file_name,
                 str(list(int(i) for i in v.px_rect)),
                 str(list(int(i) for i in v.msi_rect)),
                 str(v.spot_name),
                 str(v.TIC),
                 str(v.maxpeak),
                 str(v.rt)))
        conn.commit()
        conn.close()

    def collect_msi_img(self, target_dir):
        assert len(self.metadata) > 0, "No metadata found"
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        for k, v in self.metadata.items():
            # copy the msi image to the target directory
            shutil.copy(v.msi_img_file_path, os.path.join(target_dir, v.msi_img_file_name))


def crawl_metadata():
    """Crawl the raw data directory and return the metadata files"""
    metadata_dir = filedialog.askdirectory(title="Select the raw data directory")
    if metadata_dir:
        metadata_crawler = MetadataCrawler(metadata_dir)
        metadata_crawler.crawl()
        # ask the user to save the metadata to a sqlite database
        db_path = filedialog.asksaveasfilename(title="Save Metadata to SQLite Database",
                                               defaultextension=".db",
                                               filetypes=[("SQLite Database", "*.db")])
        if db_path:
            metadata_crawler.to_sqlite(db_path)
        # ask the user if they want to collect the msi images
        collect_msi_img = simpledialog.askstring("Collect MSI Images",
                                                 "Do you want to collect the MSI images? (y/n)")
        if collect_msi_img == "y":
            target_dir = filedialog.askdirectory(title="Select the target directory to collect MSI images")
            if target_dir:
                metadata_crawler.collect_msi_img(target_dir)
                messagebox.showinfo(
                    title="Success",
                    message=f"MSI images have been collected to {target_dir}")
            else:
                messagebox.showerror("No target directory is given")

    else:
        messagebox.showerror(
            title="Error",
            message="No raw data directory is selected")
        return


if __name__ == "__main__":
    pass



"""implement right click functionality for the application"""
import tkinter as tk
from math import sqrt
from tkinter import simpledialog, messagebox

from msiAlign.objects import MsiImage


class RightClickMenu:
    """ this is the superclass for the right-click menu"""

    def __init__(self, app):
        self.app = app
        self.menu = tk.Menu(self.app, tearoff=0)
        self._add_menu_item()
        self.clicked_item = None
        self.clicked_event = None

    def _add_menu_item(self):
        pass

    def show_menu(self, event, item=None):
        pass


class RightClickOnLine(RightClickMenu):
    """this is the right-click menu for the line"""

    def _add_menu_item(self):
        self.menu.add_command(label="Set Scale Line",
                              command=lambda: self.set_scale_line(self.clicked_item))
        self.menu.add_command(label="Set Sediment Start",
                              command=lambda: self.set_sediment_start(self.clicked_item))
        self.menu.add_command(label="Delete",
                              command=lambda: self.delete_line(self.clicked_item))

    def show_menu(self, event, item=None):
        # update the item to be right-clicked
        self.clicked_item = item
        self.clicked_event = event
        # show the menu
        self.menu.post(event.x_root, event.y_root)

    def set_scale_line(self, item):
        """tag the scale line with 'scale_line'"""
        # change the color of the scale line to blue

        self.app.items[item].label = "scale_line"
        self.app.scale_line.append(item)
        try:
            self.app.canvas.itemconfig(item, fill="blue")
        except Exception as e:
            pass

    def set_sediment_start(self, item):
        """tag the vertical line with 'sediment_start'"""
        # label the line with 'sediment_start'
        self.app.items[item].label = "sediment_start_line"
        self.app.sediment_start = item
        try:
            self.app.canvas.itemconfig(item, fill="green")
        except Exception as e:
            pass

    def delete_line(self, item):
        """delete the vertical line"""
        self.app.items[item].rm(self.app)
        # if it's the scale line, remove it from the scale_line list
        if item in self.app.scale_line:
            self.app.scale_line.remove(item)
        # if it's the sediment start line, remove it from the sediment_start
        if item == self.app.sediment_start:
            self.app.sediment_start = None


class RightClickOnImage(RightClickMenu):
    """this is the right-click menu for the image"""

    def _add_menu_item(self):

        self.menu.add_command(label="Add Label",
                              command=lambda: self.add_label(self.clicked_item))


        chg_size = tk.Menu(self.menu, tearoff=0)
        chg_size.add_command(label="x0.5", command=lambda: self.enlarge_image(self.clicked_item, 0.5))
        chg_size.add_command(label="x1.5", command=lambda: self.enlarge_image(self.clicked_item, 1.5))
        chg_size.add_command(label="x2", command=lambda: self.enlarge_image(self.clicked_item, 2))
        chg_size.add_command(label="Auto", command=lambda: self.enlarge_image(self.clicked_item, 'auto'))
        self.menu.add_cascade(label="Resize", menu=chg_size)

        self.menu.add_command(label="Use as Reference to Resize",
                              command=lambda: self.app.use_as_ref_to_resize(self.clicked_item))

        self.menu.add_command(label="Image info",
                              command=lambda: messagebox.showinfo("Image info",
                                                                  self.app.find_clicked_image(self.clicked_event)))

        self.menu.add_command(label="Unlock",
                              command=lambda: self.unlock_image(self.clicked_item))

        self.menu.add_command(label="Lock",
                              command=lambda: self.lock_image(self.clicked_item))
        self.menu.add_command(label="Delete",
                              command=lambda: self.app.items[self.clicked_item].rm(self.app))

        self.menu.add_command(label="Send to Back",
                              command=lambda: self.app.canvas.tag_lower(self.clicked_item))

    def add_label(self, item):
        """ add label to the item, unlike tag, label is not unique and can be changed and easy to understand"""
        label = simpledialog.askstring("Input", "Enter the label")
        if label:
            self.app.items[item].label = label
            # update the label column in the tree view
            self.app.tree.set(self.app.items[item].tree_master, "label", label)

    def show_menu(self, event, item=None):
        # update the item to be right-clicked
        self.clicked_item = item
        self.clicked_event = event

        if self.app.items[item].locked:
            self.menu.entryconfig("Unlock", state="normal")
            self.menu.entryconfig("Lock", state="disabled")
        else:
            self.menu.entryconfig("Unlock", state="disabled")
            self.menu.entryconfig("Lock", state="normal")
        # show the menu
        self.menu.post(event.x_root, event.y_root)

    def enlarge_image(self, item, scale_factor):
        """enlarge/shrink the image"""
        if scale_factor == 'auto':
            # make sure this is a msi image
            assert isinstance(self.app.items[item], MsiImage), messagebox.showerror("Error", "You can only auto resize the msi image")
            # assert isinstance(self.app.items[item], MsiImage), "You can only auto resize the msi image"
            assert self.app.cm_per_pixel is not None,  messagebox.showerror("Error", "Please set the scale first")
            # the real length of the slide that currently used is approximately 7.87cm
            real_size = 8
            # calculate the new size of the image
            new_width = real_size / self.app.cm_per_pixel
            scale_factor = new_width / self.app.items[item].thumbnail.width
        self.app.items[item].enlarge(scale_factor)
        self.app.canvas.itemconfig(item, image=self.app.items[item].tk_img)

    def lock_image(self, item):
        """lock the image"""
        self.app.canvas.tag_unbind(item, "<Button-1>")
        # on the left top corner, display 'locked'
        x1, y1, x2, y2 = self.app.canvas.bbox(item)
        # create a text on the canvas
        text = tk.Text(self.app.canvas, width=6, height=1)
        text.insert(tk.END, "Locked")
        text.config(state="disabled")
        self.app.canvas.create_window(x1, y1, window=text, tags=f"Locked{item}")
        self.app.items[item].lock()

    def unlock_image(self, item):
        """unlock the image"""
        self.app.canvas.tag_bind(item,
                                 "<Button-1>",
                                 lambda event: self.app.on_drag_start(item, event))
        self.app.canvas.delete(f"Locked{item}")
        self.app.items[item].unlock()


class RightClickOnTeachingPoint(RightClickMenu):
    """this is the right-click menu for the teaching point"""

    def _add_menu_item(self):
        self.menu.add_command(label="Label",
                              command=lambda: self.add_label(self.clicked_event, self.clicked_item))
        self.menu.add_separator()
        self.menu.add_command(label="Delete",
                              command=lambda: self.delete_teaching_point(self.clicked_event, self.clicked_item))

    def show_menu(self, event, item=None):
        # update the item to be right-clicked
        self.clicked_item = item
        self.clicked_event = event
        # show the menu
        self.menu.post(event.x_root, event.y_root)
        # prevent other right-click menu from showing
        return "break"

    def add_label(self, event, item):
        # find the clicked image
        clicked_image = self.app.find_clicked_image(event)
        closest_tp = None
        for k, v in clicked_image.teaching_points.items():
            # find the closest teaching point to the clicked point
            distance = sqrt((k[0] - self.app.canvas.coords(item)[0]) ** 2 +
                            (k[1] - self.app.canvas.coords(item)[1]) ** 2)
            if closest_tp is None:
                closest_tp = (k, distance)
            elif distance < closest_tp[1]:
                closest_tp = (k, distance)
        # add label to the teaching point
        while True:
            label = simpledialog.askstring("Input", "Enter the label (only integer is allowed)")
            if label is None:
                return
            try:
                label = int(label)
                break
            except ValueError:
                pass
        # append label to the teaching_points dictionary
        clicked_image.teaching_points[closest_tp[0]] = list(clicked_image.teaching_points[closest_tp[0]])
        try:
            clicked_image.teaching_points[closest_tp[0]][3] = label
        except IndexError:
            clicked_image.teaching_points[closest_tp[0]].append(label)
        clicked_image.teaching_points[closest_tp[0]] = tuple(clicked_image.teaching_points[closest_tp[0]])

    def delete_teaching_point(self, event, item):
        """delete the teaching point"""
        # find the clicked image
        clicked_image = self.app.find_clicked_image(event)
        # delete the teaching point from the teaching_points dictionary
        closest_tp = None
        for k, v in clicked_image.teaching_points.items():
            # find the closest teaching point to the clicked point
            distance = sqrt((k[0] - self.app.canvas.coords(item)[0]) ** 2 +
                            (k[1] - self.app.canvas.coords(item)[1]) ** 2)
            if closest_tp is None:
                closest_tp = (k, distance)
            elif distance < closest_tp[1]:
                closest_tp = (k, distance)
        # delete the teaching point from the teaching_points dictionary
        try:
            self.app.items[clicked_image.tag].teaching_points.pop(closest_tp[0])
        except TypeError:
            pass
        try:
            self.app.items[clicked_image.tag].teaching_points_px_coords.pop(closest_tp[0])
        except Exception as e:
            pass

        self.app.canvas.delete(item)



import numpy as np
import pandas as pd


def extract_mz(target_mz, spec, tol=0.01, min_int=10000, min_snr=0, normalization=False):
    """
    extract one m/z value from one spectrum
    :param target_mz:   the m/z value to be extracted
    :param spec:       the spectrum
    :param tol:       the tolerance for m/z values
    :param min_int:  the minimum intensity
    :param min_snr: the minimum signal-to-noise ratio
    :return:       the m/z value, intensity and signal-to-noise ratio
    """
    if normalization == "median":
        factor = np.nanmedian(spec[:, 1])
    else:
        factor = 1.0
    # sort the spectrum by intensity
    spec = spec[spec[:, 1].argsort()[::-1]]
    # find the index of the highest intensity peak within the tolerance
    spec = spec[(spec[:, 0] >= target_mz - tol / 2)
                & (spec[:, 0] <= target_mz + tol / 2)
                & (spec[:, 1] >= min_int)
                & (spec[:, 2] >= min_snr)]
    # if there is no peak within the tolerance, return nan
    if spec.shape[0] == 0:
        return np.nan, np.nan, np.nan
    else:
        return spec[0, 0], spec[0, 1]/factor, spec[0, 2]


def extract_mzs(target_mz, txt_path, tol=0.01, min_int=10000, min_snr=0, normalization=False):
    """
    this is the legacy function to read the txt file exported from Bruker DataAnalysis, and
    extract the target m/z values and intensities for all spectra from the txt file
    :param target_mz: the m/z values to be extracted, either a list of m/z values or a single m/z value
    :param txt_path: the path to the txt file
    :param tol: the tolerance for m/z values
    :param min_int: the minimum intensity
    :param min_snr: the minimum signal-to-noise ratio
    :return: a dataframe containing the target m/z values and intensities for all spectra, and the ppm error for each
    """
    # preprocessing target_mz if it's a dictionary
    if isinstance(target_mz, dict):
        mz_names = list(target_mz.keys())
        target_mz = list(target_mz.values())
    elif isinstance(target_mz, float):
        target_mz = [target_mz]
        mz_names = [0]
    else:
        mz_names = range(len(target_mz))

    # convert them to strings
    mz_names = [str(mz) for mz in mz_names]

    # read the txt file
    with open(txt_path, 'r') as f:
        lines = f.readlines()
    # get all spectra, each spectrum is a line and starts with 'R**X**Y**'
    lines = [line.strip() for line in lines]
    lines = [line for line in lines if line.startswith('R')]

    # get the spot name for each spectrum
    spot_names = [line.split(';')[0] for line in lines]
    # get the m/z values for each spectrum, it's after every two ';'
    mzs = [line.split(';')[2::3] for line in lines]
    # convert everything to floats
    mzs = [[round(float(mz), 4) for mz in mz_list] for mz_list in mzs]
    intensities = [line.split(';')[3::3] for line in lines]
    intensities = [[float(intensity) for intensity in intensity_list] for intensity_list in intensities]
    snrs = [line.split(';')[4::3] for line in lines]
    snrs = [[float(snr) for snr in snr_list] for snr_list in snrs]
    mz = np.zeros((len(spot_names), len(target_mz)))
    intensity = np.zeros((len(spot_names), len(target_mz)))
    snr = np.zeros((len(spot_names), len(target_mz)))
    for i in range(len(spot_names)):
        spec = np.array([mzs[i], intensities[i], snrs[i]]).T
        for j in range(len(target_mz)):
            mz[i, j], intensity[i, j], snr[i, j] = extract_mz(target_mz[j], spec, tol, min_int, min_snr, normalization)
    df = pd.DataFrame({'spot_name': spot_names})
    for i in range(len(target_mz)):
        df['mz_' + mz_names[i]] = mz[:, i]
        df['int_' + mz_names[i]] = intensity[:, i]
        df['snr_' + mz_names[i]] = snr[:, i]
    return df

def extract_special(txt_path, mz_range:str='full', min_int=1000, min_snr=0):
    # read the txt file
    with open(txt_path, 'r') as f:
        lines = f.readlines()
    # get all spectra, each spectrum is a line and starts with 'R**X**Y**'
    lines = [line.strip() for line in lines]
    lines = [line for line in lines if line.startswith('R')]

    # get the spot name for each spectrum
    spot_names = [line.split(';')[0] for line in lines]
    # get the m/z values for each spectrum, it's after every two ';'
    mzs = [line.split(';')[2::3] for line in lines]
    # convert everything to floats
    mzs = [[round(float(mz), 4) for mz in mz_list] for mz_list in mzs]
    intensities = [line.split(';')[3::3] for line in lines]
    intensities = [[float(intensity) for intensity in intensity_list] for intensity_list in intensities]
    snrs = [line.split(';')[4::3] for line in lines]
    snrs = [[float(snr) for snr in snr_list] for snr_list in snrs]
    result = np.zeros((len(spot_names), 3))
    for i in range(len(spot_names)):
        spec = np.array([mzs[i], intensities[i], snrs[i]]).T
        # filter out the peaks with intensity less than min_int
        spec = spec[spec[:, 1] >= min_int]
        # filter out the peaks with signal-to-noise ratio less than min_snr
        spec = spec[spec[:, 2] >= min_snr]
        if not mz_range == 'full':
            mz_range = [float(mz) for mz in mz_range.split('-')]
            spec = spec[(spec[:, 0] >= mz_range[0]) & (spec[:, 0] <= mz_range[1])]
        tic = np.sum(spec[:, 1])
        median = np.nanmedian(spec[:, 1])
        # weighted m/z
        weight_mz = np.sum(spec[:, 0] * spec[:, 1]) / tic

        result[i, 0] = tic
        result[i, 1] = median
        result[i, 2] = weight_mz
    df = pd.DataFrame({'spot_name': spot_names})
    df['tic'] = result[:, 0]
    df['median'] = result[:, 1]
    df['weight_mz'] = result[:, 2]
    return df

def extract_all(txt_path, min_int=10000, min_snr=1, peak_th=0.1, min_member=0.1, tol=10):
    # test if module 'mfe' is installed
    try:
        import mfe
        from mfe.from_txt import msi_from_txt, get_ref_peaks, create_feature_table
    except ImportError:
        raise ImportError(
            'Please install mfe first by running "pip install git+https://github.com/weimin-liu/msi_feature_extraction.git"')

    spectra = msi_from_txt(raw_txt_path=txt_path, min_int=min_int, min_snr=min_snr)

    ref = get_ref_peaks(spectra, peak_th=peak_th, min_member=min_member, tol=tol)

    feature_table = create_feature_table(spectra, ref[peak_th], tol=tol)

    return feature_table


if __name__ == "__main__":
    pass



import os
import re
import sqlite3
from tkinter import messagebox

import numpy as np
import pandas as pd

from msiAlign.parser import extract_mzs, extract_special


class DatabaseHandler:
    """Handles database interactions for the application."""

    def __init__(self, db_path):
        self.conn = sqlite3.connect(db_path)

    def close(self):
        """Closes the database connection."""
        self.conn.close()

    def get_spec_id_by_spec_file_name(self, spec_file_name):
        """Retrieves spec_id for a given spec_file_name."""
        query = 'SELECT spec_id FROM metadata WHERE spec_file_name = ?'
        return self.conn.execute(query, (spec_file_name,)).fetchall()

    def get_spec_file_name_by_export_da_name(self, export_da_name):
        """Retrieves spec_file_name for a given export_da_name."""
        query = 'SELECT spec_file_name FROM metadata WHERE export_da_name = ?'
        return self.conn.execute(query, (export_da_name,)).fetchall()

    def ensure_column_exists(self, table_name, column_name, column_type):
        """Ensures a column exists in a table; adds it if not."""
        cursor = self.conn.execute(f"PRAGMA table_info({table_name})")
        columns = [col[1] for col in cursor.fetchall()]
        if column_name not in columns:
            self.conn.execute(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}")
            self.conn.commit()

    def update_export_da_name(self, spec_file_name, export_da_name):
        """Updates the export_da_name for a given spec_file_name."""
        self.ensure_column_exists('metadata', 'export_da_name', 'TEXT')
        self.conn.execute('''
            UPDATE metadata
            SET export_da_name = ?
            WHERE spec_file_name = ?
        ''', (export_da_name, spec_file_name))
        self.conn.commit()

    def create_dataview(self):
        """Creates a view in the database for data retrieval."""
        self.conn.execute('''
            CREATE VIEW IF NOT EXISTS dataview AS
            SELECT metadata.spec_id, metadata.spec_file_name, metadata.spot_name,
                   transformation.xray_array, transformation.linescan_array
            FROM metadata
            INNER JOIN transformation ON metadata.spec_id = transformation.spec_id
        ''')

    def get_coords_by_spec_file_name(self, spec_file_name):
        """Retrieves coordinates data for a given spec_file_name."""
        query = '''
            SELECT spot_name, xray_array, linescan_array
            FROM dataview
            WHERE spec_file_name = ?
        '''
        return self.conn.execute(query, (spec_file_name,)).fetchall()

    def get_grayscale_color_by_spec_file_name(self, spec_file_name):
        query = '''
            SELECT color_values
            FROM metadata
            WHERE spec_file_name = ?
        '''
        try:
            return self.conn.execute(query, (spec_file_name,)).fetchall()
        except sqlite3.OperationalError:
            return None

    def get_tic_by_spec_file_name(self, spec_file_name):
        """Retrieves TIC data for a given spec_file_name."""
        query = '''
            SELECT TIC
            FROM metadata
            WHERE spec_file_name = ?
        '''
        try:
            return self.conn.execute(query, (spec_file_name,)).fetchall()
        except sqlite3.OperationalError:
            return None

def get_spec_file_name_from_txt(DA_txt_path):
    """Parses the spectrum file name from the txt file path."""
    return os.path.basename(DA_txt_path).replace('.txt', '')


def ensure_first_last_spotnumber_columns(db_handler):
    """
    Ensures that 'first_spot_number' and 'last_spot_number' columns exist in 'metadata'.
    Adds and populates them based on 'spot_name' if they don't.
    """
    cursor = db_handler.conn.execute("PRAGMA table_info(metadata)")
    columns = [col[1] for col in cursor.fetchall()]
    if 'first_spot_number' not in columns or 'last_spot_number' not in columns:
        if 'first_spot_number' not in columns:
            db_handler.conn.execute("ALTER TABLE metadata ADD COLUMN first_spot_number TEXT")
        if 'last_spot_number' not in columns:
            db_handler.conn.execute("ALTER TABLE metadata ADD COLUMN last_spot_number TEXT")
        db_handler.conn.commit()

        cursor = db_handler.conn.execute("SELECT spec_id, spot_name FROM metadata")
        metadata = cursor.fetchall()
        metadata_df = pd.DataFrame(metadata, columns=['spec_id', 'spot_name'])
        metadata_df['spot_list'] = metadata_df['spot_name'].str.split(',')
        metadata_df['first_spot_number'] = metadata_df['spot_list'].apply(
            lambda lst: re.findall(r'R\d+X\d+Y\d+', lst[0])[0] if lst else None)
        metadata_df['last_spot_number'] = metadata_df['spot_list'].apply(
            lambda lst: re.findall(r'R\d+X\d+Y\d+', lst[-1])[0] if lst else None)

        for _, row in metadata_df.iterrows():
            db_handler.conn.execute('''
                UPDATE metadata
                SET first_spot_number = ?, last_spot_number = ?
                WHERE spec_id = ?
            ''', (row['first_spot_number'], row['last_spot_number'], row['spec_id']))
        db_handler.conn.commit()


def extract_first_last_spotnumber(txt_path):
    """Extracts the first and last spot numbers from the txt file."""
    with open(txt_path, 'r') as f:
        lines = f.readlines()
        # Extract first spot number
        for line in lines:
            match = re.match(r'R\d+X\d+Y\d+', line)
            if match:
                first_spot_number = match.group(0)
                break
        # Extract last spot number
        for line in reversed(lines):
            match = re.match(r'R\d+X\d+Y\d+', line)
            if match:
                last_spot_number = match.group(0)
                break
    return first_spot_number, last_spot_number


def pair_txt_spec_on_first_last_spotnumber(db_handler, txt_path):
    """
    Pairs the txt file with a spectrum file name based on first and last spot numbers.
    Returns the spec_file_name if found; otherwise, None.
    """
    ensure_first_last_spotnumber_columns(db_handler)
    first_spot_number, last_spot_number = extract_first_last_spotnumber(txt_path)
    cursor = db_handler.conn.execute('''
        SELECT spec_file_name FROM metadata
        WHERE first_spot_number = ? AND last_spot_number = ?
    ''', (first_spot_number, last_spot_number))
    spec_name = cursor.fetchall()
    return spec_name[0][0] if spec_name else None


def find_spec_file_name(db_handler, DA_txt_path):
    """
    Attempts to find the spec_file_name associated with a DA_txt_path.
    Tries multiple methods and prompts the user if necessary.
    """
    spec_file_name = get_spec_file_name_from_txt(DA_txt_path)

    # Check if spec_file_name exists in the database
    if db_handler.get_spec_id_by_spec_file_name(spec_file_name):
        return spec_file_name

    # Try pairing based on first and last spot numbers
    spec_file_name = pair_txt_spec_on_first_last_spotnumber(db_handler, DA_txt_path)
    if spec_file_name:
        return spec_file_name

    # Try finding based on export_da_name
    export_da_name = os.path.basename(DA_txt_path)
    result = db_handler.get_spec_file_name_by_export_da_name(export_da_name)
    if len(result) == 1:
        return result[0][0]

    # Prompt user for spec_file_name
    if messagebox.askyesno("Error", "The spectrum file name does not exist in the database. "
                                    "Do you want to manually type the spectrum file name?"):
        while True:
            spec_file_name = input("Please enter the spectrum file name: ")
            result = db_handler.get_spec_id_by_spec_file_name(spec_file_name)
            if len(result) == 1:
                db_handler.update_export_da_name(spec_file_name, export_da_name)
                return spec_file_name
            else:
                messagebox.showerror("Error", "The spectrum file name does not exist in the database or is not unique")
    else:
        return None


def get_mz_int_depth(DA_txt_path, db_path, target_cmpds=None, tol=0.01, min_snr=1, min_int=10000,
                     normalization=False) -> pd.DataFrame:
    """
    Extracts m/z intensities and depth information from a DA exported text file,
    combining it with metadata from the database.
    """
    db_handler = DatabaseHandler(db_path)
    try:
        spec_file_name = find_spec_file_name(db_handler, DA_txt_path)
        if spec_file_name is None:
            messagebox.showerror("Error", "The spectrum file name could not be determined.")
            return pd.DataFrame()

        # Extract target compounds from the DA_txt_path
        if target_cmpds is None:
            df = extract_special(DA_txt_path, mz_range='full', min_snr=min_snr)
        else:
            df = extract_mzs(target_cmpds, DA_txt_path, tol=tol, min_snr=min_snr,
                             min_int=min_int, normalization=normalization)

        # Create data view and retrieve coordinates
        db_handler.create_dataview()
        coords_result = db_handler.get_coords_by_spec_file_name(spec_file_name)
        coords_result_color = db_handler.get_grayscale_color_by_spec_file_name(spec_file_name)
        coords_result_tic = db_handler.get_tic_by_spec_file_name(spec_file_name)

        if not coords_result:
            messagebox.showerror("Error", "The spectrum file name does not exist in the database. "
                                          "Ensure the DA export file name matches the spectrum file name in the database.")
            return pd.DataFrame()

        coords = coords_result[0]
        spot_names = [re.findall(r'R\d+X\d+Y\d+', name)[0] for name in coords[0].split(',')]
        spot_names_df = pd.DataFrame(spot_names, columns=['spot_name'])

        xray_array = np.frombuffer(coords[1], dtype=np.float64).reshape(-1, 2)
        xray_array_df = pd.DataFrame(xray_array, columns=['px', 'py'])

        linescan_array = np.frombuffer(coords[2], dtype=np.float64).reshape(-1, 2)
        linescan_array_df = pd.DataFrame(linescan_array[:, 0], columns=['d'])

        if coords_result_color:
            try:
                color_values = np.frombuffer(coords_result_color[0][0], dtype=np.uint8)
                color_values = color_values.reshape(-1, 1)
                color_values_df = pd.DataFrame(color_values, columns=['color_values'])
            except TypeError:
                color_values_df = pd.DataFrame(np.nan, index=np.arange(len(spot_names)), columns=['color_values'])
        else:
            color_values_df = pd.DataFrame(np.nan, index=np.arange(len(spot_names)), columns=['color_values'])

        if coords_result_tic:
            try:
                coords_result_tic = coords_result_tic[0][0]
                coords_result_tic = eval(coords_result_tic)
                tic_df = pd.DataFrame(coords_result_tic, columns=['TIC'])
                tic_df = tic_df.astype({'TIC': float})
            except TypeError:
                tic_df = pd.DataFrame(np.nan, index=np.arange(len(spot_names)), columns=['TIC'])
        else:
            tic_df = pd.DataFrame(np.nan, index=np.arange(len(spot_names)), columns=['TIC'])

        # Combine dataframes
        coords_df = pd.concat([spot_names_df, xray_array_df, linescan_array_df,color_values_df, tic_df], axis=1)
        df = pd.merge(coords_df, df, on='spot_name')

        if normalization == "TIC" or normalization == "tic":
            # normalization all 'int_' columns by TIC
            int_cols = [col for col in df.columns if col.startswith('int_')]
            df[int_cols] = df[int_cols].div(df['TIC'], axis=0)

        return df

    finally:
        db_handler.close()


def get_chunks(depth, horizon_size, min_n_samples=10,dynamic=False,res=200,max_extra_row=5):
    """
    Divides the depth array into chunks based on horizon size and minimum number of samples.
    Returns a list of (start_index, end_index) tuples representing chunks.
    """
    depth = np.array(depth)
    if not np.all(np.diff(depth) >= 0):
        raise ValueError("The depth array is not sorted")

    start_index = 0
    chunks = []
    n = len(depth)

    while start_index < n:
        end_value = depth[start_index] + horizon_size
        current_index = start_index
        while current_index < n and depth[current_index] <= end_value:
            current_index += 1
        if current_index - start_index >= min_n_samples:
            chunks.append((start_index, current_index))
        else:
            # if the chunk is too small, depend on dynamic to decide if we should add more rows
            if dynamic:
                # if the chunk is too small, add more rows until the chunk is larger than min_n_samples
                retry = 0
                while current_index - start_index < min_n_samples and current_index < n and retry < max_extra_row:
                    end_value = depth[current_index] + res
                    retry += 1
                    while current_index < n and depth[current_index] <= end_value:
                        current_index += 1
                if current_index - start_index >= min_n_samples:
                    chunks.append((start_index, current_index))
            else:
                # if the chunk is too small, ignore it
                pass

        start_index = current_index


    return chunks


def to_1d(df, chunks, how: str) -> pd.DataFrame:
    """
    Aggregates data in df over the given chunks using the specified method.
    Returns a DataFrame of aggregated values for each chunk.
    """
    int_cols = [col for col in df.columns if 'int' in col or 'tic' in col
                or 'median' in col or 'weight_mz' in col]
    df_1d_list = []

    for chunk in chunks:
        start_index, end_index = chunk
        data = df.iloc[start_index:end_index]
        int_data = data[int_cols]
        valid_counts = int_data.count()
        valid_mask = valid_counts >= 10

        if how == 'mean':
            v = int_data.mean()
            v[~valid_mask] = 0
            df_1d_list.append(v)
        elif how == 'std':
            v = int_data.std()
            v[~valid_mask] = 0
            df_1d_list.append(v)
        elif how == 'sum':
            v = int_data.sum()
            v[~valid_mask] = 0
            df_1d_list.append(v)
        elif how == 'median':
            v = int_data.median()
            v[~valid_mask] = 0
            df_1d_list.append(v)
        elif how == 'sumall':
            v = int_data.sum().sum()
            df_1d_list.append(v)

        else:
            # Evaluate custom expression
            try:
                safe_dict = {'data': data}
                v = eval(how, {"__builtins__": None}, safe_dict)
                df_1d_list.append(v)
            except Exception as e:
                messagebox.showerror(title="Error", message=f"Error evaluating expression: {e}")
                return pd.DataFrame()

    df_1d = pd.DataFrame(df_1d_list)
    return df_1d



def get_msi_depth_profile_from_gui(exported_txt_path, sqlite_db_path, target_cmpds, how, spot_method,dynamic,dyn_res,dyn_max_retry, tol, min_snr, min_int,
                               min_n_samples,
                               horizon_size, save_path, save_path_1d, additional_params,
                                   **kwargs):
    # conver all values to float
    tol = float(tol)
    min_snr = float(min_snr)
    min_int = float(min_int)
    min_n_samples = int(min_n_samples)
    spot_method = spot_method
    dynamic = bool(dynamic)
    dyn_res = float(dyn_res) / 10000  # convert to cm
    dyn_max_retry = int(dyn_max_retry)

    additional_params = additional_params
    additional_params = {param.split(':')[0]: param.split(':')[1] for param in additional_params.split(';') if param.strip() != ''}

    if 'normalization' in additional_params:
        normalization = additional_params['normalization']
    else:
        normalization = False

    horizon_size = float(horizon_size) / 10000  # convert to cm
    # convert target_cmpds string "name1:mz1;name2:mz2" to a dictionary
    # handle possible trailing semicolons gracefully
    cmpd_pairs = [c for c in target_cmpds.split(';') if c.strip()]
    target_cmpds = dict(pair.split(':') for pair in cmpd_pairs)
    # make sure the values are floats
    target_cmpds = {name: float(mz) for name, mz in target_cmpds.items()}
    # get exported_txt_path, seperated by ';' and trim the last ';' if there is one
    exported_txt_path = exported_txt_path.split(';')
    if exported_txt_path[-1] == '':
        exported_txt_path = exported_txt_path[:-1]
    # find the first existing path
    df_1d_list = []
    for path in exported_txt_path:
        if os.path.exists(path):
            single_exported_txt_path = path
            df = get_mz_int_depth(single_exported_txt_path, sqlite_db_path, target_cmpds, tol=tol, min_snr=min_snr,
                                  min_int=min_int, normalization=normalization)
            df = df.sort_values(by='d')

            # save the dataframe
            # append save_path with index if there are multiple exported_txt_path
            if len(exported_txt_path) > 1:
                _save_path = save_path.replace('.csv', f'_{exported_txt_path.index(single_exported_txt_path)}.csv')
            else:
                _save_path = save_path
            df.to_csv(_save_path, index=False)

            # this part record the missing depth for correct plotting
            if not dynamic:
                fake_chunks = get_chunks(df['d'], horizon_size, min_n_samples=0)
                fake_depth_1d = to_1d(df, fake_chunks, "data['d'].mean()")
                fake_depth_1d.columns = ['d']

            # replace the 0s in the int columns with np.nan
            int_cols = [col for col in df.columns if 'int' in col ]
            df[int_cols] = df[int_cols].replace(0, np.nan)
            if spot_method == 'all':
                # in all mode, drop the int columns where any of the compounds is zero
                df = df.dropna(subset=int_cols, how='any')
            elif spot_method == 'any':
                # in any mode, drop the int columns where all the compounds are zero
                df = df.dropna(subset=int_cols, how='all')
            else:
                # in custom mode, drop the column by spot_method
                custom_cmpd_list = [s for s in spot_method.split(';') if s.strip()]
                custom_cmpd_list = ['int_'+custom_cmpd for custom_cmpd in custom_cmpd_list]
                df = df.dropna(subset=custom_cmpd_list, how='any')
            # fill the nan with 0
            df[int_cols] = df[int_cols].fillna(0)
            df = df.sort_values(by='d')

            chunks = get_chunks(df['d'], horizon_size, min_n_samples=min_n_samples,dynamic=dynamic,res=dyn_res,max_extra_row=dyn_max_retry)
            if len(chunks) == 0:
                df_1d = pd.DataFrame(columns=['d (cm)', 'horizon_count','horizon_len (cm)', 'slide', 'result'])
            else:
                # get the mean depth of each chunk
                depth_1d = to_1d(df, chunks, "data['d'].mean()")
                depth_1d_min = to_1d(df, chunks, "data['d'].min()")
                depth_1d_max = to_1d(df, chunks, "data['d'].max()")
                depth_1d_size = depth_1d_max - depth_1d_min
                ratio_1d = to_1d(df, chunks, how)
                horizon_count = [chunk[1] - chunk[0] for chunk in chunks]
                df_1d = pd.DataFrame({'d (cm)': depth_1d.iloc[:, 0],
                                      'horizon_count': horizon_count,
                                      'horizon_len (cm)': depth_1d_size.iloc[:, 0],
                                      'slide': [os.path.basename(single_exported_txt_path)] * len(depth_1d),
                                      'result': ratio_1d.iloc[:, 0]})

                try:
                    if not dynamic:
                        # add the missing depth to the result
                        # for each of the depth in fake_depth_1d, if it is not in df_1d +- horizon_size, add it to df_1d
                        for idx, row in fake_depth_1d.iterrows():
                            if not ((row['d'] - horizon_size < df_1d['d (cm)']) & (df_1d['d (cm)'] < row['d'] + horizon_size)).any():
                                df_1d = pd.concat(
                                    [df_1d,
                                     pd.DataFrame(
                                         {'d (cm)': [row['d']],
                                          'horizon_count': np.nan,
                                          'horizon_len (cm)': np.nan,
                                          'slide': [os.path.basename(single_exported_txt_path)],
                                          'result': np.nan})],
                                    axis=0, ignore_index=True)
                # if anything goes wrong, just pass, as it is not critical when the missing depth is not added
                except Exception as e:
                    print(e)
                    pass

            df_1d = df_1d.sort_values(by='d (cm)')

            # save the 1d depth profile
            if len(exported_txt_path) > 1:
                _save_path_1d = save_path_1d.replace('.csv',
                                                     f'_{exported_txt_path.index(single_exported_txt_path)}.csv')
            else:
                _save_path_1d = save_path_1d
            df_1d.to_csv(_save_path_1d, index=False)
            df_1d_list.append(df_1d)

    # Concatenate all 1D depth profiles if multiple files
    if len(df_1d_list) > 1:
        all_df_1d = pd.concat(df_1d_list, axis=0, ignore_index=True)
        all_df_1d.to_csv(save_path_1d.replace('.csv', '_all.csv'), index=False)
    else:
        all_df_1d = df_1d_list[0]

    show_tk_message = kwargs.get('show_tk_message', True)
    if show_tk_message:
        # create a tkinter messagebox to show the user it's done and add an ok button to close the window
        messagebox.showinfo(title="Done", message="The downcore profile has been successfully created")

    return_df_1d = kwargs.get('return_df_1d', False)
    if return_df_1d:
        return all_df_1d


def get_xrf_depth_profile_from_gui(exported_csv_path, how,
                               min_n_samples,
                               horizon_size, save_path_1d):
    # conver all values to float
    min_n_samples = int(min_n_samples)
    horizon_size = float(horizon_size) / 10000 # convert to cm
    # parse all the target elements from how: Al/Ca, Ca/Ti, etc., ignoring trailing semicolons
    how = [h for h in how.split(';') if h.strip()]
    # get exported_csv_path, seperated by ';' and trim the last ';' if there is one
    exported_csv_path = exported_csv_path.split(';')
    if exported_csv_path[-1] == '':
        exported_csv_path = exported_csv_path[:-1]
    df_list = []
    # find the first existing path
    for path in exported_csv_path:
        if os.path.exists(path):
            single_exported_csv_path = path
            df = pd.read_csv(single_exported_csv_path)
            df = df.dropna()
            # use the mask
            try:
                df = df[df['mask']]
            except Exception as e:
                print(e)
                return
            try:
                df = df.sort_values(by='d')
            except KeyError:
                messagebox.showerror("Error", "The exported csv file does not contain a 'd' column.")
                return
            chunks = get_chunks(df['d'], horizon_size, min_n_samples=min_n_samples)
            # get the mean depth of each chunk
            depth_1d = to_1d(df, chunks, "data['d'].mean()")
            ratio_1ds = []
            for r_how in how:
                if '/' in r_how:
                    e0 = r_how.split('/')[0]
                    e1 = r_how.split('/')[1]
                    ratio_1d = to_1d(df, chunks, f"data['{e0}'].sum()/data['{e1}'].sum()")
                    ratio_1ds.append(ratio_1d)
                else:
                    ratio_1d = to_1d(df, chunks, f"data['{r_how}'].mean()")
                    ratio_1ds.append(ratio_1d)
            horizon_count = [chunk[1] - chunk[0] for chunk in chunks]
            df_1d = pd.DataFrame({'d': depth_1d.iloc[:, 0],
                                    'horizon_count': horizon_count})
            for idx, ratio_1d in enumerate(ratio_1ds):
                df_1d[f'{how[idx]}'] = ratio_1d.iloc[:, 0]

            # save the 1d depth profile
            if len(exported_csv_path) > 1:
                _save_path_1d = save_path_1d.replace('.csv',
                                                     f'_{exported_csv_path.index(single_exported_csv_path)}.csv')
            else:
                _save_path_1d = save_path_1d
            df_1d.to_csv(_save_path_1d, index=False)
            df_list.append(df_1d)

    if len(df_list) > 1:
        all_df_1d = pd.concat(df_list, axis=0, ignore_index=True)
        all_df_1d.to_csv(save_path_1d.replace('.csv', '_all.csv'), index=False)

    # create a tkinter messagebox to show the user it's done and add an ok button to close the window
    messagebox.showinfo("Done", "The downcore profile has been successfully created")


if __name__ == "__main__":
    pass


import re
import sqlite3

import numpy as np


def store_blob_info(conn, var_name, dtype, shape):
    """store the blob info to the sqlite database"""
    c = conn.cursor()
    try:
        # check if 'blob_info' table exists
        c.execute('SELECT * FROM blob_info')
    except sqlite3.OperationalError:
        # create a blob_info table
        c.execute('CREATE TABLE blob_info (var_name TEXT, dtype TEXT, shape TEXT)')
        conn.commit()
    # store the blob info to the blob_info table
    c.execute('INSERT INTO blob_info (var_name, dtype, shape) VALUES (?, ?, ?)',
              (var_name, str(dtype), str(shape)))
    conn.commit()


def get_msi_rect_from_imaginginfo(xml_file, return_spot_name=False):
    with open(xml_file, 'r') as f:
        xml = f.read()
    # get all the value between <spotName> and </spotName>
    spot_name = re.findall(r'<spotName>(.*?)</spotName>', xml)
    tic = re.findall(r'<tic>(.*?)</tic>', xml)
    maxpeak = re.findall(r'<maxpeak>(.*?)</maxpeak>', xml)
    rt = re.findall(r'<minutes>(.*?)</minutes>', xml)
    if return_spot_name:
        spot_name_str = spot_name.copy()
    # parse the spot name to get x and y
    spot_name = [re.findall(r'X(\d+)Y(\d+)', s) for s in spot_name]
    # flatten the list
    spot_name = [item for sublist in spot_name for item in sublist]
    # convert to int
    spot_name = [(int(x), int(y)) for x, y in spot_name]
    # convert to array
    spot_name = np.array(spot_name)
    # get the coordinates of the four corners
    x_min, y_min = spot_name.min(axis=0)
    x_max, y_max = spot_name.max(axis=0)

    if return_spot_name:
        return [x_min, y_min, x_max, y_max], spot_name_str, tic, maxpeak, rt
    else:
        return [x_min, y_min, x_max, y_max]


def get_image_file_from_mis(mis_file):
    with open(mis_file, 'r') as f:
        mis = f.read()
    # get the value between <ImageFile> and </ImageFile>
    image_file = re.findall(r'<ImageFile>(.*?)</ImageFile>', mis)
    return image_file[0]


def get_px_rect_from_mis(mis_file):
    with open(mis_file, 'r') as f:
        mis = f.read()
    # get all the value between <Point> and </Point>
    roi_bound = re.findall(r'<Point>(.*?)</Point>', mis)
    roi_bound = [s.split(',') for s in roi_bound]
    roi_bound = [(int(x), int(y)) for x, y in roi_bound]
    roi_bound = np.array(roi_bound)
    # get a rectangle shape from the points
    x_min, y_min = roi_bound.min(axis=0)
    x_max, y_max = roi_bound.max(axis=0)
    return [x_min, y_min, x_max, y_max]


class CorSolver:
    """
    A class that takes tie points from two coordinate system (i.e., one from Fleximaging, the other from
    xray measurement) and solve the transformation between these two systems.
    """

    def __init__(self):
        self.translation_vector = None
        self.transformation_matrix = None

    def _reset(self):
        if hasattr(self, 'transformation_matrix'):
            del self.transformation_matrix
            del self.translation_vector

    def fit(self, src_tri, dst_tri):
        """
        Parameters:
        --------
            src_tri: coordinates of the source triangle in the source coordinate system
            dst_tri:  coordinates of the target triangle in the target coordinate system
        """
        self._reset()
        return self.partial_fit(src_tri, dst_tri)

    def partial_fit(self, src_tri, dst_tri):
        """
        solve the affine transformation matrix between FlexImage coordinates and X_ray coordinates
        https://stackoverflow.com/questions/56166088/how-to-find-affine-transformation-matrix-between-two-sets-of-3d-points
        """
        len_source = len(src_tri)
        basis = np.vstack([np.transpose(src_tri), np.ones(len_source)])
        diagonal = 1.0 / np.linalg.det(basis)

        def entry(r_val, d_val):
            return np.linalg.det(np.delete(np.vstack([r_val, basis]), (d_val + 1), axis=0))

        m_matrix = [[(-1) ** i * diagonal * entry(R, i) for i in range(len_source)] for R in
                    np.transpose(dst_tri)]
        # pylint: disable=unbalanced-tuple-unpacking
        a_matrix, t_val = np.hsplit(np.array(m_matrix), [len_source - 1])
        t_val = np.transpose(t_val)[0]
        self.transformation_matrix = a_matrix
        self.translation_vector = t_val
        return self

    def transform(self, src_coordinate):
        """
        Parameters:
        --------
            src_coordinate: the source coordinates that needs be transformed
        """
        dst_coordinate = src_coordinate.dot(self.transformation_matrix.T) + self.translation_vector
        return dst_coordinate


if __name__ == "__main__":
    pass



import tkinter as tk
from tkinter import filedialog
from tkinter.ttk import Combobox

from msiAlign.to1d import get_msi_depth_profile_from_gui, get_xrf_depth_profile_from_gui


def validate_input(target_cmpds, spot_method):
    """
    Validate the input
    :param target_cmpds: target compounds
    :param spot_method: method to combine spots
    :return:
    """
    if spot_method.lower() == 'all' or spot_method.lower() == 'any':
        return True
    else:
        cmpds = target_cmpds.split(';')
        cmpds = [c.split(':')[0] for c in cmpds]
        spot_method = spot_method.split(';')
        for s in spot_method:
            if s not in cmpds:
                return False
        return True


def calc_depth_profile_combined():
    """Calculate the depth profile (combined mode)"""
    window = tk.Toplevel()
    window.title("Calc Downcore Profile")
    widgets = {}
    row_index = 0

    # Mode selector
    tk.Label(window, text="Mode:").grid(row=row_index, column=0, sticky='nsew')
    mode_selector = Combobox(window, values=['MSI', 'XRF'])
    mode_selector.current(0)
    mode_selector.grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    # MSI mode widgets
    widgets['exported_txt_label'] = tk.Label(window, text="Exported txt path(s):")
    widgets['exported_txt_entry'] = tk.Entry(window)
    widgets['exported_txt_button'] = tk.Button(window, text="Select", command=lambda: widgets['exported_txt_entry'].insert(
        tk.END,
        ';'.join(filedialog.askopenfilenames(title='Select exported txt', filetypes=[("Plain text files", "*.txt")])) + ';'
    ))

    widgets['exported_txt_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['exported_txt_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['exported_txt_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    widgets['sqlite_db_label'] = tk.Label(window, text="Sqlite db path:")
    widgets['sqlite_db_entry'] = tk.Entry(window)
    widgets['sqlite_db_button'] = tk.Button(window, text="Select", command=lambda: widgets['sqlite_db_entry'].insert(
        tk.END,
        filedialog.askopenfilename(title='Select sqlite db', filetypes=[("SQLite files", "*.db")])
    ))

    widgets['sqlite_db_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['sqlite_db_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['sqlite_db_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    widgets['target_cmpds_label'] = tk.Label(window, text="Target cmpds:")
    widgets['target_cmpds_combo'] = Combobox(window, values=(
        "name1:mz1;name2:mz2",
        "GDGT0:1324.3046;GDGT5:1314.2264",
        "ALK37_2:553.5319;ALK37_3:551.5162"
    ))
    widgets['target_cmpds_combo'].current(0)

    widgets['target_cmpds_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['target_cmpds_combo'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['how_label'] = tk.Label(window, text="How:")
    widgets['how_entry'] = Combobox(window, values=(
        "data['int_name1'].sum() / (data['int_name1'].sum() + data['int_name2'].sum())",
        "sumall"
    ))
    widgets['how_entry'].current(0)

    widgets['how_button'] = tk.Button(window, text="Generate", command=lambda: [
        widgets['how_entry'].delete(0, tk.END),
        widgets['how_entry'].insert(
            tk.END,
            f"data['int_{widgets['target_cmpds_combo'].get().split(';')[0].split(':')[0]}'].sum() / "
            f"(data['int_{widgets['target_cmpds_combo'].get().split(';')[0].split(':')[0]}'].sum() +"
            f"data['int_{widgets['target_cmpds_combo'].get().split(';')[1].split(':')[0]}'].sum())"
        )
    ])


    widgets['how_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['how_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['how_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    widgets['method_label'] = tk.Label(window, text="Spot method:")
    widgets['method_entry'] = Combobox(window, values=(
        "all",
        "any",
        "other:type in the cmpds, sep by ';'"))
    widgets['method_entry'].current(0)
    widgets['method_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['method_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['method_dynamic_label'] = tk.Label(window, text="Dynamic spots:")
    # add a checkbox to enable dynamic spots or not
    widgets['method_dynamic_var'] = tk.IntVar()
    widgets['method_dynamic_checkbox'] = tk.Checkbutton(window, variable=widgets['method_dynamic_var'])
    widgets['method_dynamic_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['method_dynamic_checkbox'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    # add dynamic spots parameters
    widgets['dyn_res_label'] = tk.Label(window, text="MSI res (μm):")
    widgets['dyn_res_entry'] = tk.Entry(window)
    widgets['dyn_res_entry'].insert(tk.END, "Only for dynamic spots, default is 200")
    widgets['dyn_res_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['dyn_res_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['dyn_max_retry'] = tk.Label(window, text="Max extra rows:")
    widgets['dyn_max_retry_entry'] = tk.Entry(window)
    widgets['dyn_max_retry_entry'].insert(tk.END, "Only for dynamic spots, default is 5")
    widgets['dyn_max_retry'].grid(row=row_index, column=0, sticky='nsew')
    widgets['dyn_max_retry_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1


    widgets['tol_label'] = tk.Label(window, text="Tol (Da, full window):")
    widgets['tol_entry'] = tk.Entry(window)
    widgets['tol_entry'].insert(tk.END, "0.01")

    widgets['tol_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['tol_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['min_snr_label'] = tk.Label(window, text="Min snr:")
    widgets['min_snr_entry'] = tk.Entry(window)
    widgets['min_snr_entry'].insert(tk.END, "1")

    widgets['min_snr_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['min_snr_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['min_int_label'] = tk.Label(window, text="Min int:")
    widgets['min_int_entry'] = tk.Entry(window)
    widgets['min_int_entry'].insert(tk.END, "10000")

    widgets['min_int_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['min_int_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['save_2d_label'] = tk.Label(window, text="Save 2D to:")
    widgets['save_2d_entry'] = tk.Entry(window)
    widgets['save_2d_entry'].insert(tk.END, "2D_res.csv")
    widgets['save_2d_button'] = tk.Button(window, text="Select", command=lambda: [
        widgets['save_2d_entry'].delete(0, tk.END),
        widgets['save_2d_entry'].insert(
            tk.END,
            filedialog.asksaveasfilename(title="Save 2d result as",
                                         defaultextension=".csv",
                                         filetypes=[("CSV file", '*.csv')])
        )
    ])

    widgets['save_2d_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['save_2d_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['save_2d_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    # XRF mode widgets
    widgets['transformed_csv_label'] = tk.Label(window, text="Transformed csv path(s):")
    widgets['transformed_csv_entry'] = tk.Entry(window)
    widgets['transformed_csv_button'] = tk.Button(window, text="Select", command=lambda: widgets['transformed_csv_entry'].insert(
        tk.END,
        ';'.join(filedialog.askopenfilenames(title='Select transformed csv', filetypes=[("CSV files", "*.csv")])) + ';'
    ))

    widgets['transformed_csv_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['transformed_csv_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['transformed_csv_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    widgets['ratios_label'] = tk.Label(window, text="Ratios (sep by ';'):")
    widgets['ratios_entry'] = tk.Entry(window)
    widgets['ratios_entry'].insert(tk.END, "Ca/Ti;Fe/Ti")

    widgets['ratios_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['ratios_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    # Common widgets
    widgets['min_n_samples_label'] = tk.Label(window, text="Min-n-spots/horizon:")
    widgets['min_n_samples_entry'] = tk.Entry(window)
    widgets['min_n_samples_entry'].insert(tk.END, "10")

    widgets['min_n_samples_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['min_n_samples_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['horizon_size_label'] = tk.Label(window, text="Horizon size (μm):")
    widgets['horizon_size_entry'] = tk.Entry(window)
    widgets['horizon_size_entry'].insert(tk.END, "500")

    widgets['horizon_size_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['horizon_size_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    widgets['save_1d_label'] = tk.Label(window, text="Save 1D to:")
    widgets['save_1d_entry'] = tk.Entry(window)
    widgets['save_1d_entry'].insert(tk.END, "1D_res.csv")
    widgets['save_1d_button'] = tk.Button(window, text="Select", command=lambda: [
        widgets['save_1d_entry'].delete(0, tk.END),
        widgets['save_1d_entry'].insert(
            tk.END,
            filedialog.asksaveasfilename(title="Save 1d result as",
                                         defaultextension=".csv",
                                         filetypes=[("CSV file", '*.csv')])
        )
    ])

    widgets['save_1d_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['save_1d_entry'].grid(row=row_index, column=1, sticky='nsew')
    widgets['save_1d_button'].grid(row=row_index, column=2, sticky='nsew')
    row_index += 1

    # add a widget to input additional parameters for the calculation
    widgets['additional_params_label'] = tk.Label(window, text="Additional params:")
    widgets['additional_params_entry'] = tk.Entry(window)
    widgets['additional_params_entry'].insert(tk.END, "normalization:none;")
    widgets['additional_params_label'].grid(row=row_index, column=0, sticky='nsew')
    widgets['additional_params_entry'].grid(row=row_index, column=1, sticky='nsew')
    row_index += 1

    # Start button
    widgets['start_button'] = tk.Button(window, text="Start")
    widgets['start_button'].grid(row=row_index, column=0, columnspan=3, sticky='nsew')

    # Column configurations
    for i in range(3):
        window.grid_columnconfigure(i, weight=1 if i == 1 else 0)

    # Function to update widget visibility based on mode
    def update_widgets():
        mode = mode_selector.get()
        if mode == 'MSI':
            # Show MSI mode widgets
            widgets['exported_txt_label'].grid()
            widgets['exported_txt_entry'].grid()
            widgets['exported_txt_button'].grid()
            widgets['sqlite_db_label'].grid()
            widgets['sqlite_db_entry'].grid()
            widgets['sqlite_db_button'].grid()
            widgets['target_cmpds_label'].grid()
            widgets['target_cmpds_combo'].grid()
            widgets['how_label'].grid()
            widgets['how_entry'].grid()
            widgets['how_button'].grid()
            widgets['method_label'].grid()
            widgets['method_entry'].grid()
            # Dynamic spots
            widgets['method_dynamic_label'].grid()
            widgets['method_dynamic_checkbox'].grid()
            widgets['dyn_res_label'].grid()
            widgets['dyn_res_entry'].grid()
            widgets['dyn_max_retry'].grid()
            widgets['dyn_max_retry_entry'].grid()


            widgets['tol_label'].grid()
            widgets['tol_entry'].grid()
            widgets['min_snr_label'].grid()
            widgets['min_snr_entry'].grid()
            widgets['min_int_label'].grid()
            widgets['min_int_entry'].grid()
            widgets['save_2d_label'].grid()
            widgets['save_2d_entry'].grid()
            widgets['save_2d_button'].grid()
            widgets['additional_params_label'].grid()
            widgets['additional_params_entry'].grid()

            # Hide XRF mode widgets
            widgets['transformed_csv_label'].grid_remove()
            widgets['transformed_csv_entry'].grid_remove()
            widgets['transformed_csv_button'].grid_remove()
            widgets['ratios_label'].grid_remove()
            widgets['ratios_entry'].grid_remove()
        elif mode == 'XRF':
            # Hide MSI mode widgets
            widgets['exported_txt_label'].grid_remove()
            widgets['exported_txt_entry'].grid_remove()
            widgets['exported_txt_button'].grid_remove()
            widgets['sqlite_db_label'].grid_remove()
            widgets['sqlite_db_entry'].grid_remove()
            widgets['sqlite_db_button'].grid_remove()
            widgets['target_cmpds_label'].grid_remove()
            widgets['target_cmpds_combo'].grid_remove()
            widgets['how_label'].grid_remove()
            widgets['how_entry'].grid_remove()
            widgets['how_button'].grid_remove()

            widgets['method_dynamic_label'].grid_remove()
            widgets['method_dynamic_checkbox'].grid_remove()
            widgets['dyn_res_label'].grid_remove()
            widgets['dyn_res_entry'].grid_remove()
            widgets['dyn_max_retry'].grid_remove()
            widgets['dyn_max_retry_entry'].grid_remove()


            widgets['method_label'].grid_remove()
            widgets['method_entry'].grid_remove()
            widgets['tol_label'].grid_remove()
            widgets['tol_entry'].grid_remove()
            widgets['min_snr_label'].grid_remove()
            widgets['min_snr_entry'].grid_remove()
            widgets['min_int_label'].grid_remove()
            widgets['min_int_entry'].grid_remove()
            widgets['save_2d_label'].grid_remove()
            widgets['save_2d_entry'].grid_remove()
            widgets['save_2d_button'].grid_remove()
            widgets['additional_params_label'].grid_remove()
            widgets['additional_params_entry'].grid_remove()

            # Show XRF mode widgets
            widgets['transformed_csv_label'].grid()
            widgets['transformed_csv_entry'].grid()
            widgets['transformed_csv_button'].grid()
            widgets['ratios_label'].grid()
            widgets['ratios_entry'].grid()

    # Function to start the calculation
    def start_calculation():
        mode = mode_selector.get()
        # Common parameters
        min_n_samples = widgets['min_n_samples_entry'].get()
        horizon_size = widgets['horizon_size_entry'].get()
        save_1d_path = widgets['save_1d_entry'].get()
        if mode == 'MSI':
            # Get values from MSI mode widgets
            exported_txt_path = widgets['exported_txt_entry'].get()
            sqlite_db_path = widgets['sqlite_db_entry'].get()
            target_cmpds = widgets['target_cmpds_combo'].get()
            how = widgets['how_entry'].get()
            spot_method = widgets['method_entry'].get()
            additional_params = widgets['additional_params_entry'].get()

            # Dynamic spots
            # if dynamic spots is enabled, get the values as boolean
            dynamic = widgets['method_dynamic_var'].get()
            dynamic = True if dynamic == 1 else False
            if dynamic:
                dyn_res = widgets['dyn_res_entry'].get()
                dyn_max_retry = widgets['dyn_max_retry_entry'].get()
                # try converting to float, if failed, use default values
                try:
                    dyn_res = float(dyn_res)
                except ValueError:
                    dyn_res = 200
                try:
                    dyn_max_retry = int(dyn_max_retry)
                except ValueError:
                    dyn_max_retry = 5
            else:
                dyn_res = 200
                dyn_max_retry = 5

            tol = widgets['tol_entry'].get()
            min_snr = widgets['min_snr_entry'].get()
            min_int = widgets['min_int_entry'].get()
            save_2d_path = widgets['save_2d_entry'].get()

            if not validate_input(
                target_cmpds=target_cmpds,
                spot_method=spot_method,
            ):
                tk.messagebox.showerror("Error", "Invalid input, please make sure the cmpds listed in 'Method' are valid")
                return

            # Call the processing function
            get_msi_depth_profile_from_gui(
                exported_txt_path, sqlite_db_path, target_cmpds, how, spot_method, dynamic, dyn_res, dyn_max_retry, tol, min_snr, min_int,
                min_n_samples, horizon_size, save_2d_path, save_1d_path, additional_params
            )
        else:
            # XRF mode
            exported_txt_path = widgets['transformed_csv_entry'].get()
            how = widgets['ratios_entry'].get()
            get_xrf_depth_profile_from_gui(
                exported_txt_path, how,
                min_n_samples, horizon_size, save_1d_path
            )


    # Bindings and initializations
    mode_selector.bind("<<ComboboxSelected>>", lambda event: update_widgets())
    widgets['start_button'].config(command=start_calculation)
    update_widgets()


import json
import os
import re
import sqlite3
import sys
import tkinter as tk
from tkinter import filedialog
from tkinter import simpledialog, messagebox

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from msiAlign.func import CorSolver
from msiAlign.menubar import MenuBar
from msiAlign.objects import LoadedImage, VerticalLine, MsiImage, TeachableImage
from msiAlign.rclick import RightClickOnLine, RightClickOnImage, RightClickOnTeachingPoint


class MainApplication(tk.Tk):
    def __init__(self):
        super().__init__()
        self.geometry("1200x800")
        self.canvas = None
        self.right_click_on_tp = None
        self.right_click_on_image = None
        self.right_click_on_line = None
        self.title('Ctrl+O to add images, Shift+click to add teaching points, Ctrl+click to add rulers')
        self.items = {}
        self.create_canvas()
        self.xrf_folder = None
        self.database_path = None
        self.pair_tp_str = None

        self.scale_line = []
        self.sediment_start = None

        self.cm_per_pixel = None
        self.save_attrs = [
            'cm_per_pixel', 'database_path', 'pair_tp_str', 'sediment_start', 'scale_line'
        ]
        self.calculation_handler = CalculationHandler(self)
        self.dev_ops_handler = DevOpsHandler(self)
        self.xrf_handler = XRFHandler(self)
        self.menu = MenuBar(self)

        self.create_right_click_op()

    def create_right_click_op(self):
        self.right_click_on_line = RightClickOnLine(self)
        self.right_click_on_image = RightClickOnImage(self)
        self.right_click_on_tp = RightClickOnTeachingPoint(self)

    def create_canvas(self):
        canvas_width = 1000
        canvas_height = 1000
        # Create a frame for the canvas and scrollbars
        canvas_frame = tk.Frame(self)
        canvas_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.canvas = tk.Canvas(canvas_frame, width=canvas_width, height=canvas_height, bg='white',
                                scrollregion=(0, 0, 5000, 5000))
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        # Create horizontal and vertical scrollbars
        h_scroll = tk.Scrollbar(self.canvas, orient='horizontal', command=self.canvas.xview)
        h_scroll.pack(side=tk.BOTTOM, fill='x')
        v_scroll = tk.Scrollbar(canvas_frame, orient='vertical', command=self.canvas.yview)
        v_scroll.pack(side=tk.RIGHT, fill='y')
        # bind the mousewheel event to the canvas
        # Configure the canvas to use the scrollbars
        self.canvas.configure(xscrollcommand=h_scroll.set, yscrollcommand=v_scroll.set)
        self.canvas.bind_all("<MouseWheel>", self.on_mousewheel)
        self.canvas.bind_all("<Shift-MouseWheel>", self.horizontal_mousewheel)

    def on_mousewheel(self, event):
        try:
            # For windows and MacOS
            # if it's macos
            if sys.platform == "darwin":
                self.canvas.yview_scroll(event.delta, "units")
            # else it's windows
            elif sys.platform == "win32":
                self.canvas.yview_scroll(int(-1 * (event.delta / 120)), "units")
        except AttributeError:
            raise AttributeError("The mousewheel event is not supported on this platform")

    def horizontal_mousewheel(self, event):
        try:
            # For windows and MacOS
            self.canvas.xview_scroll(event.delta, "units")
        except AttributeError:
            raise AttributeError("The mousewheel event is not supported on this platform")

    def on_drag_start(self, item, event):
        """Function to handle dragging"""
        # Get the coordinates of the image
        x1, y1, x2, y2 = self.canvas.bbox(item)

        # create a temporary variable to store the old width
        self._old_width = self.items[item].thumbnail.width

        # calculate the offset
        _drag_offset_x = self.canvas.canvasx(event.x) - x1
        _drag_offset_y = self.canvas.canvasy(event.y) - y1
        # Create a rectangle around the image
        self.canvas.create_rectangle(x1, y1, x2, y2, outline="blue", tags="rect")

        # Calculate the corner threshold as a percentage of the image's width and height
        corner_threshold_x = (x2 - x1) * 0.1
        corner_threshold_y = (y2 - y1) * 0.1

        # if the mouse is near the bottom-right corner, start resizing
        if (abs(x2 - self.canvas.canvasx(event.x)) < corner_threshold_x
                and abs(y2 - self.canvas.canvasy(event.y)) < corner_threshold_y):
            self.canvas.bind("<B1-Motion>", lambda e: self.on_resize(e, item))
            self.canvas.bind("<ButtonRelease-1>", lambda e: self.on_resize_stop(e, item))

        else:
            self.canvas.bind("<B1-Motion>", lambda e: self.on_drag_move(e, item, _drag_offset_x, _drag_offset_y))
            self.canvas.bind("<ButtonRelease-1>", lambda e: self.on_drag_stop(e, item))

    def on_drag_move(self, event, item, _drag_offset_x, _drag_offset_y):
        """move the item to the new position"""
        x, y = self.canvas.canvasx(event.x) - _drag_offset_x, self.canvas.canvasy(event.y) - _drag_offset_y
        self.canvas.coords(item, x, y)

    def on_drag_stop(self, event, item):
        """Stop dragging the image"""
        # record the origin offset of the image
        offset_x = self.canvas.coords(item)[0] - self.items[item].origin[0]
        offset_y = self.canvas.coords(item)[1] - self.items[item].origin[1]
        # record the new position of the image
        x, y = self.canvas.coords(item)
        self.items[item].origin = (x, y)
        # remove the rectangle from the canvas
        self.canvas.delete('rect')
        self.canvas.unbind("<B1-Motion>")
        self.canvas.unbind("<ButtonRelease-1>")
        # update the teaching points
        if isinstance(self.items[item], TeachableImage):
            self.items[item].update_teaching_points(self, offset_x, offset_y)

    def on_resize(self, event, item):
        """Resize the image based on mouse position"""
        x1, y1, x2, y2 = self.canvas.bbox(item)
        new_width = self.canvas.canvasx(event.x) - x1
        # Prevent the image from being too small
        new_width = max(new_width, 100)
        # get the original PIL image
        image = self.items[item].thumbnail
        # calculate the aspect ratio of the original image
        aspect_ratio = image.width / image.height
        # calculate the new height based on the aspect ratio
        new_height = new_width / aspect_ratio
        # resize the image by creating a new image with the new dimensions
        self.items[item].resize((new_width, new_height))

        self.canvas.itemconfig(item, image=self.items[item].tk_img)
        # TODO: an error will be raised here if there are teaching points on the image, no need
        # to panic. This is because the teaching points have the same tag as the image, so they
        # are also be configured, but they cannot be configured with the image, so the error is raised

    def on_resize_stop(self, event, item):
        """Stop resizing the image"""
        self.resizing = False
        # remove the rectangle from the canvas
        self.canvas.delete('rect')
        self.canvas.unbind("<B1-Motion>")
        self.canvas.unbind("<ButtonRelease-1>")
        if isinstance(self.items[item], TeachableImage):
            self.items[item].update_teaching_points_on_resize(self, self.items[item].origin,
                                                              self.items[item].thumbnail.width / self._old_width)

    def add_vertical_line(self, event):
        """draw a ruler on the canvas when ctrl-left-click is pressed, and calculate the scale"""
        vl = VerticalLine((self.canvas.canvasx(event.x), 0))
        self.items[vl.tag] = vl
        vl.create_on_canvas(self)

        # calculate the pixel distance between this ruler and sediment_start
        if self.sediment_start is not None and self.cm_per_pixel is not None:
            pixel_distance = self.canvas.coords(self.sediment_start)[0] - self.canvas.canvasx(event.x)
            real_distance = pixel_distance * self.cm_per_pixel
            vl.add_depth_text(self, f"{abs(real_distance):.2f}")

    def find_clicked_image(self, event):
        # TODO: could be optimized by using addtag_withtag when creating the images and the items
        clicked_images = []
        for k, v in self.items.items():
            if isinstance(v, LoadedImage):
                x1, y1, x2, y2 = self.canvas.bbox(v.tag)
                if x1 <= self.canvas.canvasx(event.x) <= x2 and y1 <= self.canvas.canvasy(event.y) <= y2:
                    clicked_images.append(v)

        if len(clicked_images) == 1:
            return clicked_images[0]
        elif len(clicked_images) > 1:
            # when images are overlapping, find the front image
            # first call find_all to get the current order of the images
            current_order = self.canvas.find_all()
            # sort the clicked images based on the current order
            clicked_images = sorted(clicked_images,
                                    key=lambda x: current_order.index(self.canvas.find_withtag(x.tag)[0]))
            return clicked_images[-1]
        else:
            return None

    def bind_events_to_loaded_images(self, loaded_image):
        """Bind events to the loaded images"""
        self.canvas.tag_bind(f"{loaded_image.tag}", "<Button-1>",
                             lambda e, item=f"{loaded_image.tag}": self.on_drag_start(item, e))
        # bind ctrl-left-click to add a ruler
        self.canvas.tag_bind(f"{loaded_image.tag}", "<Control-Button-1>", self.add_vertical_line)
        # bind shift-left-click to add a teaching point
        self.canvas.tag_bind(f"{loaded_image.tag}", "<Shift-Button-1>",
                             lambda e: self.add_teaching_point(e))
        # bind right-click event to the image
        if sys.platform == "darwin":
            self.canvas.tag_bind(f"{loaded_image.tag}",
                                 "<Button-2>",
                                 lambda e, item=f"{loaded_image.tag}": self.right_click_on_image.show_menu(e, item))
        elif sys.platform == "win32":
            self.canvas.tag_bind(f"{loaded_image.tag}",
                                 "<Button-3>",
                                 lambda e, item=f"{loaded_image.tag}": self.right_click_on_image.show_menu(e, item))
        else:
            raise ValueError("The platform is not supported")

    def add_tp_labels(self):
        """Add labels for all the teaching points"""
        label_idx = 0
        for k, v in self.items.items():
            try:
                for i, tp in v.teaching_points.items():
                    # append label to the teaching_points dictionary
                    v.teaching_points[i] = list(v.teaching_points[i])
                    try:
                        v.teaching_points[i][3] = label_idx
                    except IndexError:
                        v.teaching_points[i].append(label_idx)
                    v.teaching_points[i] = tuple(v.teaching_points[i])

                    try:
                        # pass the label to teaching point px coords
                        v.teaching_points_px_coords[i] = list(v.teaching_points_px_coords[i])
                        try:
                            v.teaching_points_px_coords[i][3] = label_idx
                        except IndexError:
                            v.teaching_points_px_coords[i].append(label_idx)
                        v.teaching_points_px_coords[i] = tuple(v.teaching_points_px_coords[i])
                    except AttributeError:
                        pass
                    except IndexError:
                        messagebox.showerror("Error", "Something went wrong when adding labels to teaching points")
                    label_idx += 1

            except AttributeError:
                pass
        # show the labels
        self.show_tp_labels()

    def show_tp_labels(self):
        """ display the labels of the teaching points on the canvas"""
        for k, v in self.items.items():
            if isinstance(v, TeachableImage):
                v.show_tp_labels(self)

    def hide_tp_labels(self):
        """delete all the text items on the canvas labeled 'tp_labels'"""
        self.canvas.delete('tp_labels')

    def fill_tps_str(self):
        try:
            return self.pair_tp_str
        except AttributeError:
            return ""

    def bind_events_to_vertical_lines(self, vertical_line):
        """Bind events to the vertical lines"""
        if sys.platform == "darwin":
            self.canvas.tag_bind(f"{vertical_line.tag}",
                                 "<Button-2>",
                                 lambda e, item=f"{vertical_line.tag}": self.right_click_on_line.show_menu(e, item))
        elif sys.platform == "win32":
            self.canvas.tag_bind(f"{vertical_line.tag}",
                                 "<Button-3>",
                                 lambda e, item=f"{vertical_line.tag}": self.right_click_on_line.show_menu(e, item))
        else:
            raise ValueError("The platform is not supported")
        # bind shift-left-click to add a teaching point
        self.canvas.tag_bind(f"{vertical_line.tag}", "<Shift-Button-1>", self.add_teaching_point)

    def add_teaching_point(self, event):
        """Add a teaching point to the canvas"""
        clicked_image = self.find_clicked_image(event)
        if not isinstance(clicked_image, TeachableImage) and not isinstance(clicked_image, MsiImage):
            messagebox.showerror("Wrong image", "Click a Teachable image to add a teaching point")
            return
        if clicked_image is not None:
            clicked_image.add_teaching_point(event, self)

    def add_metadata(self):
        """Add metadata to the app"""
        # note that the metadata needed to be added to the app after all msi images are loaded
        # ask the user if all the msi images are loaded
        userchoice = messagebox.askokcancel(
            "Add Metadata",
            "All the MSI images need to be loaded before adding metadata, continue adding metadata?",
            icon='warning'
        )
        if not userchoice:
            return
        file_path = filedialog.askopenfilename(title="Select a database file", filetypes=[("SQLite files", "*.db")])
        if file_path:
            self.database_path = file_path
        else:
            messagebox.showerror("No file path is given")
        # connect to the sqlite database
        import sqlite3
        conn = sqlite3.connect(file_path)
        c = conn.cursor()
        # get the image name, px_rect, and msi_rect
        c.execute('SELECT msi_img_file_name, px_rect, msi_rect FROM metadata')
        data = c.fetchall()
        count = 0
        for row in data:
            im_name, px_rect, msi_rect = row
            # attach the metadata to the corresponding image
            try:
                self.items[im_name.replace(' ','_')].px_rect = eval(px_rect)
                self.items[im_name.replace(' ','_')].msi_rect = eval(msi_rect)
                count += 1
            except KeyError:
                pass
        conn.close()
        # create a popup window to show the process is done, and ok button to close the window
        messagebox.showinfo("Success", f"{count} images have been added with metadata")

    def use_as_ref_to_resize(self, item):
        """use the selected image as the reference to resize other images"""
        ref_width = self.items[item].thumbnail.width
        for k, v in self.items.items():
            if isinstance(v, MsiImage):
                scale_factor = ref_width / v.thumbnail.width
                self.items[k].enlarge(scale_factor)
                try:
                    self.canvas.itemconfig(k, image=self.items[k].tk_img)
                except tk.TclError:
                    pass

    def save_layout(self, event=None):
        """Save the layout of the canvas"""
        self.save(layout_only=True)

    def save(self, event=None, layout_only=False):
        """Save the current state of the canvas"""
        file_path = filedialog.asksaveasfilename(title="Save workspace",
                                                 defaultextension=".json",
                                                 filetypes=[("JSON", "*.json")])
        if not file_path:
            return  # User cancelled the save dialog
        data_to_save = {}

        # Save the attributes automatically
        for attr in self.save_attrs:
            data_to_save[attr] = getattr(self, attr, None)

        # Save items
        data_to_save["items"] = []
        for k, v in self.items.items():
            if not layout_only or not isinstance(v, MsiImage):
                data_to_save["items"].append(v.to_json())

        with open(file_path, "w") as f:
            json.dump(data_to_save, f)

    def load(self, event=None):
        """Load the state of the canvas"""
        file_path = filedialog.askopenfilename(title="Select a workspace file", filetypes=[("JSON files", "*.json")])
        if not file_path:
            return  # User cancelled the load dialog
        with open(file_path, "r") as f:
            data = json.load(f)
            # Reset the canvas
            self.dev_ops_handler.reset()

            # Load the attributes automatically
            for attr in self.save_attrs:
                setattr(self, attr, data.get(attr, None))

            # Recreate canvas text for cm_per_pixel
            if self.cm_per_pixel is not None:
                text = tk.Text(self.canvas, height=1, width=20)
                text.insert(tk.END, f"1cm = {int(1 / self.cm_per_pixel)} pixel")
                text.config(state="disabled")
                self.canvas.create_window(100, 100, window=text, tags="cm_per_px_text")

            # Load items
            for item in data.get("items", []):
                if "MsiImage" in item["type"]:
                    loaded_image = MsiImage.from_json(item, self)
                    self.items[loaded_image.tag] = loaded_image
                elif "TeachableImage" in item["type"]:
                    loaded_image = TeachableImage.from_json(item, self)
                    self.items[loaded_image.tag] = loaded_image
                elif item["type"] == "VerticalLine":
                    vertical_line = VerticalLine.from_json(item, self)
                    self.items[vertical_line.tag] = vertical_line
                    self.bind_events_to_vertical_lines(vertical_line)

            # Reconfigure scale lines and sediment start if they exist
            if self.scale_line:
                # Ensure that scale_line contains items
                if len(self.scale_line) >= 2:
                    self.canvas.itemconfig(self.scale_line[0], fill="blue")
                    self.canvas.itemconfig(self.scale_line[1], fill="blue")
            if self.sediment_start:
                self.canvas.itemconfig(self.sediment_start, fill="green")


    def find_wildcard(self, wildcard):
        """find the tag with the wildcard"""
        items = self.canvas.find_all()
        matched_items = []
        for item in items:
            if wildcard in self.canvas.gettags(item)[0]:
                matched_items.append(self.canvas.gettags(item))
        return matched_items

    def main(self):
        self.mainloop()

class DevOpsHandler:
    def __init__(self, app):
        self.app = app

    def set_tp_size(self):
        """set the size of the teaching points"""
        size = simpledialog.askinteger("Input", "Enter the size of the teaching points", initialvalue=5)
        size = size if size is not None else 5
        for k, v in self.app.items.items():
            if isinstance(v, TeachableImage):
                v.tp_size = size

    def reset(self):
        userchoice=tk.messagebox.showwarning("Warning", "This will reset the canvas, are you sure?")
        if userchoice == 'yes':
            # reset the canvas
            self.app.canvas.delete("all")
            self.app.items = {}
            self.app.cm_per_pixel = None
            self.app.database_path = None
            self.app.pair_tp_str = None
            self.app.scale_line = []
            self.app.sediment_start = None
            self.app.pair_tp_str = None
            self.app.solvers_xray = {}
            self.app.solvers_depth = {}
        else:
            pass

    def export_tps(self):
        """Export the teaching points to a json file"""
        file_path = filedialog.asksaveasfilename(title="Export Teaching Points",
                                                 defaultextension=".json",
                                                 filetypes=[("JSON files", "*.json")])
        if file_path:
            data_to_save = "img;x;y;d\n"
            for k, v in self.app.items.items():
                if hasattr(v, "teaching_points"):
                    for k, tp in v.teaching_points.items():
                        data_to_save += f"{v.tag};{tp[0]};{tp[1]};{tp[2]}\n"
            with open(file_path, "w") as f:
                f.write(data_to_save)
            messagebox.showinfo("Export Teaching Points", f"Teaching points have been exported to {file_path}")
        else:
            messagebox.showinfo("Export Teaching Points", "No file path is given")
            return

    def reset_tp(self):
        """
        Reset the teaching points
        """
        # remove all the teaching points from the canvas
        for k, v in self.app.items.items():
            if isinstance(v, TeachableImage):
                if v.teaching_points is not None:
                    v.teaching_points = {}
                if hasattr(v, "teaching_points_updated"):
                    v.teaching_points_updated = False
        # hard remove all the teaching points oval from the canvas with tag including 'tp_'
        try:
            # list all tags
            tps = self.app.find_wildcard('tp_')
            for tp in tps:
                try:
                    self.app.canvas.delete(tp[0])
                except IndexError:
                    self.app.canvas.delete(tp)
        except AttributeError:
            pass
        messagebox.showinfo("Done", "All the teaching points are removed")

        # clear the tree view
        try:
            self.app.tree.delete(*self.app.tree.get_children())
        except AttributeError:
            pass

    def calc_depth_for_all_tps(self):
        """calculate the depth for all the teaching points"""
        if self.app.sediment_start is None:
            messagebox.showerror("Error", "No sediment start is found")
            return
        if self.app.cm_per_pixel is None:
            messagebox.showerror("Error", "No cm_per_pixel is found")
            return
        user_choice1 = messagebox.askyesno("Warning",
                                           "This will overwrite the current depth of the teaching points,"
                                           " are you sure?")
        user_choice2 = messagebox.askyesno("Warning",
                                           "If you have moved the xray images after setting the teaching points, "
                                           "the depth will be incorrect, are you sure?")

        if user_choice1 and user_choice2:
            for k, v in self.app.items.items():
                if isinstance(v, TeachableImage):
                    try:
                        for px_coords, values in v.teaching_points.items():
                            depth = abs(self.app.canvas.coords(self.app.sediment_start)[0] - px_coords[0]) * self.app.cm_per_pixel
                            _tmp = list(values)
                            _tmp[2] = depth
                            v.teaching_points[px_coords] = tuple(_tmp)
                    except Exception as e:
                        messagebox.showerror("Error", f"Error: {e}")
                    try:
                        for px_coords, values in v.teaching_points_px_coords.items():
                            depth = abs(self.app.canvas.coords(self.app.sediment_start)[0] - px_coords[0]) * self.app.cm_per_pixel
                            _tmp = list(values)
                            _tmp[2] = depth
                            v.teaching_points_px_coords[px_coords] = tuple(_tmp)
                    except AttributeError:
                        pass

            messagebox.showinfo("Done", "The depth for all the teaching points are calculated")

    def lock_all(self):
        # invoke lock_image method for all the images
        for k, v in self.app.items.items():
            if isinstance(v, LoadedImage):
                self.app.right_click_on_image.lock_image(k)
        messagebox.showinfo("Done", "All the images are locked")

class CalculationHandler:
    def __init__(self, app):
        self.app = app

    def calc_cm_per_px(self):
        # get the two vertical scale lines
        if len(self.app.scale_line) < 2:
            messagebox.showerror("Error", "You need to draw two vertical lines to calculate the scale")
            raise ValueError("You need to draw two vertical lines to calculate the scale")
        elif len(self.app.scale_line) > 2:
            messagebox.showerror("Error", f"You have drawn more than two vertical lines:{self.app.scale_line} ")
            raise ValueError("You have drawn more than two vertical lines")
        else:
            pixel_distance = abs(
                self.app.canvas.coords(self.app.scale_line[1])[0] - self.app.canvas.coords(self.app.scale_line[0])[0])
            real_distance = simpledialog.askfloat("Real Distance", "Real Distance (cm):")
            self.app.cm_per_pixel = real_distance / pixel_distance
            text = tk.Text(self.app.canvas, height=1, width=20)
            text.insert(tk.END, f"1cm = {int(1 / self.app.cm_per_pixel)} pixel")
            text.config(state="disabled")
            self.app.canvas.create_window(100, 100, window=text, tags="cm_per_px_text")

    def calc_msi_machine_coordinate(self):
        for k, v in self.app.items.items():
            if isinstance(v, MsiImage):
                try:
                    v.update_tp_coords(self.app.database_path)
                except AttributeError:
                    messagebox.showerror("Error", "No database path is found")
                    return

    def machine_to_real_world(self):
        """apply the transformation to the msi teaching points"""
        # ask for the sqlite file to read the metadata
        if self.app.database_path is None:
            self.app.add_metadata()
        # connect to the sqlite database
        file_path = self.app.database_path
        if file_path:
            # connect to the sqlite database
            conn = sqlite3.connect(file_path)
            c = conn.cursor()
            try:
                # check if the transformation table exists
                c.execute('SELECT * FROM transformation')
                # purge the transformation table
                c.execute('DELETE FROM transformation')
                # delete the transformation table
                c.execute('DROP TABLE transformation')
                conn.commit()
            except sqlite3.OperationalError:
                pass
            # create a transformation table with metadata(spec_id) as the reference key
            try:
                c.execute(
                    'CREATE TABLE transformation (spec_id INTEGER, msi_img_file_name TEXT, spot_array BLOB, xray_array BLOB, linescan_array BLOB, FOREIGN KEY(spec_id) REFERENCES metadata(spec_id))')
                conn.commit()
            except sqlite3.OperationalError:
                pass
            # read all the spotname from metadata table and convert them to array
            c.execute('SELECT spec_id, msi_img_file_name, spot_name FROM metadata')
            data = c.fetchall()
            assert len(data) > 0, "No data is found in the metadata table"
            for row in data:
                spec_id, im_name, spot_name = row
                spec_id = int(spec_id)
                spot_name = eval(spot_name)
                # apply the transformation to the spot_name
                spot_name = [re.findall(r'X(\d+)Y(\d+)', s) for s in spot_name]
                # flatten the list
                spot_name = [item for sublist in spot_name for item in sublist]
                # convert to an array
                spot_name = np.array(spot_name)
                # conver the spotname to int
                spot_name = spot_name.astype(int)
                # write the spotnames to the transformation table as a blob
                c.execute('INSERT INTO transformation (spec_id, msi_img_file_name, spot_array) VALUES (?, ?, ?)',
                          (spec_id, im_name, spot_name.tobytes()))
                conn.commit()
            c.execute('SELECT spec_id, msi_img_file_name, spot_array FROM transformation')
            data = c.fetchall()
            for row in data:
                spec_id, im_name, spot_array = row
                spec_id = int(spec_id)
                spot_array = np.frombuffer(spot_array, dtype=int).reshape(-1, 2)
                # apply the transformation to the spot_array
                if im_name in self.app.solvers_xray.keys() or im_name.replace(' ', '_') in self.app.solvers_xray.keys():
                    xray_array = self.app.solvers_xray[im_name.replace(' ','_')].transform(spot_array)
                    xray_array_dtype = xray_array.dtype
                    # store_blob_info(conn, 'xray_array', xray_array_dtype, xray_array.shape)
                    xray_array_shape = xray_array.shape
                    linescan_array = self.app.solvers_depth[im_name.replace(' ', '_')].transform(spot_array)
                    # line_scan_dtype = linescan_array.dtype
                    # linescan_array_shape = linescan_array.shape
                    # store_blob_info(conn, 'linescan_array', line_scan_dtype, linescan_array_shape)
                    c.execute('UPDATE transformation SET xray_array = ? WHERE spec_id = ?',
                              (xray_array.tobytes(), spec_id))
                    c.execute('UPDATE transformation SET linescan_array = ? WHERE spec_id = ?',
                              (linescan_array.tobytes(), spec_id))
            conn.commit()
            conn.close()
        else:
            messagebox.showerror("No file path is given")

        # create a popup window to show the process is done, and ok button to close the window
        messagebox.showinfo("Done", "The transformation is applied to the MSI coords")

    def pair_tps(self, str1, msi=False, xrf=False):
        self.app.pair_tp_str = str1
        # remove the leading and trailing white spaces
        str1 = str1.strip()
        # convert the input string to a list of tuples
        str1 = str1.split('\n')
        str1 = [s.split(' ') for s in str1]

        str1 = [[int(x) for x in s] for s in str1]
        paired_tps = str1
        # get all teaching points from xray:
        for k, v in self.app.items.items():
            if isinstance(v, TeachableImage) and not isinstance(v, MsiImage):
                xray_tps = v.label_indexed_teaching_points
                break
        # get all their labels
        xray_tp_labels = list(xray_tps.keys())
        # convert self.app.paired_tps to a dictionary using xray_tp_labels as the values
        paired_tps_dict = {}
        for i, v in enumerate(paired_tps):
            if v[0] in xray_tp_labels:
                paired_tps_dict[v[1]] = v[0]
            elif v[1] in xray_tp_labels:
                paired_tps_dict[v[0]] = v[1]
            else:
                continue

        self.app.solvers_xray = {}
        self.app.solvers_depth = {}

        for k, v in self.app.items.items():
            if isinstance(v, MsiImage) or isinstance(v, TeachableImage):
                try:
                    msi_tps = v.label_indexed_teaching_points
                    # find the teaching points that have the paired label in xray_tps
                    partial_xray_tps = {}
                    for msi_tp_label in msi_tps.keys():
                        if msi_tp_label in paired_tps_dict.keys():
                            partial_xray_tps[msi_tp_label] = xray_tps[paired_tps_dict[msi_tp_label]]
                    self.app.solvers_xray[k] = CorSolver()
                    self.app.solvers_xray[k].fit(np.array(list(msi_tps.values()))[:, 0:2],
                                             np.array(list(partial_xray_tps.values()))[:, 0:2])
                    self.app.solvers_depth[k] = CorSolver()
                    self.app.solvers_depth[k].fit(np.array(list(msi_tps.values()))[:, 0:2],
                                              np.array(list(partial_xray_tps.values()))[:, [2, 1]])
                except Exception as e:
                    print(e)
                    pass
        if msi:
            self.machine_to_real_world()

        if xrf:
            self.app.xrf_handler.prepare_for_xrf()

    def calc_msi(self):
        # test if metadata is added
        if self.app.database_path is None:
            # create a messagebox to show that the metadata is not added yet
            userchoice = messagebox.askokcancel(
                "Metadata is not added yet",
                "Do you want to add metadata now?",
                icon='warning'
            )
            if userchoice:
                self.app.add_metadata()
            else:
                return

        # calculate the MSI machine coordinate

        self.calc_msi_machine_coordinate()
        # create a messagebox to ask if the user wants to pair the teaching points automatically or manually
        self.app.menu.pair_tps_ui(msi=True)


class XRFHandler:
    def __init__(self, app):
        self.app = app
        self.xrf_folder = None
        self.elements = None

    def set_xrf_folder(self):
        """set the folder to store the xrf images"""
        folder_path = filedialog.askdirectory(title="Select a folder that contains all the XRF data")
        if folder_path:
            self.xrf_folder = folder_path
            self.read_all_elements()
        else:
            messagebox.showerror("No folder path is given")


    def read_all_elements(self):
        """read all the elements from the xrf images"""
        if self.xrf_folder is None:
            self.set_xrf_folder()
        if self.xrf_folder:
            # list all the folders in the xrf folder
            xrf_folders = [f for f in os.listdir(self.xrf_folder) if os.path.isdir(os.path.join(self.xrf_folder, f))]
            self.elements = {}
            for a_folder in xrf_folders:
                # get all the xrf images in the folder (.txt files without 'Video' in the name)
                xrf_files = [f for f in os.listdir(os.path.join(
                    self.xrf_folder, a_folder
                )) if f.endswith('.txt')]
                if len(xrf_files) == 0:
                    # ignore empty folders
                    continue
                element = {}
                # read all the elements from the xrf images
                # find the changing parts and the common parts of all names
                common_part = os.path.commonprefix(xrf_files)
                # safely ignore the folder if there is no common part
                if common_part == '':
                    messagebox.showwarning(
                        title='Warning',
                        message=f'No common part found in the txt files in the folder {a_folder}, skipping the folder'
                    )
                    continue
                changing_parts = [f.replace(common_part, '').replace('.txt', '') for f in xrf_files]
                for i, f in enumerate(xrf_files):
                    try:
                        element[changing_parts[i]] = pd.read_csv(os.path.join(self.xrf_folder,a_folder, f), sep=';',header=None)
                    except Exception as e:
                        print(e)
                        pass # sometimes there are corrupted txt files in the xrf folder, skip those
                # convert the elements to a dataframe, with x and y and the element names as the columns
                for k, v in element.items():
                    # convert the wide format to long format
                    v = v.reset_index()
                    v = pd.melt(v,id_vars=['index'], var_name='y', value_name=k)
                    v = v.rename(columns={'index': 'x'})
                    element[k] = v
                # concatenate all the elements to a single dataframe, with x,y as the common keys
                self.elements[a_folder] = pd.concat(list(element.values()), axis=1).loc[:,~pd.concat(list(element.values()), axis=1).columns.duplicated()]

    def mask_xrf_data(self,by='Fe'):
        """mask the xrf data by the element Fe"""
        if self.elements is None:
            self.read_all_elements()
        # mask the xrf data by the element Fe
        # do a k means clustering on the Fe data
        from sklearn.cluster import KMeans
        kmeans = KMeans(n_clusters=2)
        for a_folder in self.elements.keys():
            kmeans.fit(self.elements[a_folder][by].values.reshape(-1, 1))
            # get the cluster centers
            cluster_centers = kmeans.cluster_centers_
            # get the cluster labels
            cluster_labels = kmeans.labels_
            # get the cluster with the higher mean
            cluster = np.argmax(cluster_centers)

            # mask the xrf data by the cluster
            self.elements[a_folder]['mask'] = cluster_labels == cluster
            # save the mask as a image in the folder,use a discrete colormap
            plt.imshow(self.elements[a_folder].pivot(
                index='x', columns='y', values='mask'
            ), cmap='viridis')
            plt.axis('off')
            plt.savefig(os.path.join(self.xrf_folder, a_folder, 'mask.png'))
            plt.close()

    def transform_xrf_data(self):
        for a_folder in self.elements.keys():
            # loop through the solver_xray keys
            for k, v in self.app.solvers_depth.items():
                # test if the key image is in the folder
                if k in [i.replace(' ','_') for i in os.listdir(os.path.join(self.xrf_folder,a_folder))]:
                    self.elements[a_folder]['d'] = v.transform(self.elements[a_folder][['y', 'x']].values)[:,0]
            # save the transformed data to the folder
            self.elements[a_folder].to_csv(os.path.join(self.xrf_folder, a_folder, 'transformed.csv'), index=False)

    def prepare_for_xrf(self):
        """prepare the app for XRF data"""
        # ask the user for the by element
        by = simpledialog.askstring("Input", "Enter the element to mask the XRF data by", initialvalue='Video')
        by = by.strip()
        # add the xrf data to the app
        self.read_all_elements()
        # mask the xrf data by the element Fe
        # pop up a window to ask for the element to mask the xrf data

        self.mask_xrf_data(by=by)
        # transform the xrf data to the real world
        self.transform_xrf_data()


def main():
    app = MainApplication()
    app.main()


if __name__ == "__main__":
    main()



from shiny import App, render, ui, reactive
import matplotlib.pyplot as plt
import webbrowser
import threading
import time

from msiAlign.to1d import get_msi_depth_profile_from_gui
import os

class MSIDownCoreProfileApp:
    def __init__(self):
        self.ui = self.create_ui()
        self.server = self._server
        self.app = App(ui=self.ui, server=self.server)


    def plot_1d(self, df_1d):
        # plot the result
        df_1d = df_1d.sort_values(by='d (cm)')
        fig, ax = plt.subplots()
        ax.plot(
            df_1d['d (cm)'],
            df_1d['result']
        )
        ax.set_xlabel('Depth (cm)')
        ax.set_ylabel('Result')
        return fig



    def create_ui(self):
        return ui.page_fluid(
            ui.h1("MSI Down Core Profile Calculator"),
            ui.card(
                ui.card_header("Path to txts and metadata"),
                ui.input_text(
                    "txt_files",
                    label="Path to txt files",
                    placeholder="path/to/txt/files0;path/to/txt/files1;path/to/txt/files2",
                    width='100%'
                ),
                ui.input_text(
                    "metadata",
                    label="Path to metadata",
                    placeholder="path/to/metadata.db",
                    width = '100%'
                ),
                ui.input_text(
                    "save2d",
                    label="Path to save 2D profile",
                    placeholder="path/to/save/2d/profile",
                    width = '100%'
                ),
                ui.input_text(
                    "save1d",
                    label="Path to save 1D profile",
                    placeholder="path/to/save/1d/profile",
                    width = '100%'
                ),
            ),
            ui.card(
                ui.card_header("Spectral parameters"),
                ui.input_text(
                    "target_cmpds",
                    label="Target compounds",
                    placeholder="name1:mz1;name2:mz2",
                    width = '100%'
                ),
                ui.input_text(
                    "how",
                    label="How to calculate",
                    placeholder="data['int_name1'].sum() / (data['int_name1'].sum() + data['int_name2'].sum())",
                    width = '100%'
                ),
                ui.input_numeric(
                    "tol",
                    label="Tolerance (Da, full width)",
                    value=0.01,
                ),
                ui.input_numeric(
                    "min_snr",
                    label="Minimum SNR",
                    value=1,
                ),
                ui.input_numeric(
                    "min_int",
                    label="Minimum intensity",
                    value=1e5,
                ),
                ui.input_checkbox(
                    "normalization",
                    label="TIC Normalization",
                    value=False,
                )
            ),
            ui.card(
                ui.card_header("Spot parameters"),
                ui.input_checkbox_group(
                    "spot_method",
                    label="Spot method (choose only one)",
                    choices=['all', 'any', 'other'],
                    selected = 'all'
                ),
                ui.input_checkbox(
                    "dynamic_spots",
                    label="Dynamic spots",
                    value=False,
                ),
                ui.input_numeric(
                    "MSI_res",
                    label="MSI resolution (um)",
                    value=200,
                ),
                ui.input_numeric(
                    "max_extra_rows",
                    label="Max extra rows",
                    value = 5,
                ),
            ),

            ui.card(
                ui.card_header("Horizon parameters"),
                ui.input_numeric(
                    "min_n_spots",
                    label="Minimum number of spots per horizon",
                    value = 10
                ),
                ui.input_numeric(
                    "horizon_size",
                    label="Horizon size (um)",
                    value=500
                )

            ),
            ui.input_action_button(
                "calculate",
                label="Calculate",
                class_="btn-primary"
            ),

            ui.output_plot(
                "downcore_profile_fig",
            )
        )

    def _server(self, input, output, session):

        downcore_fig = reactive.value(None)

        @reactive.effect
        @reactive.event(input.calculate)
        def _():
            exported_txt_path = input.txt_files()
            sqlite_db_path = input.metadata()
            save_2d_path = input.save2d()
            save_1d_path = input.save1d()

            target_cmpds = input.target_cmpds()
            how = input.how()
            try:
                spot_method = input.spot_method()[0]
            except IndexError:
                spot_method = 'all'
            dynamic = input.dynamic_spots()
            tol = input.tol()
            min_snr = input.min_snr()
            min_int = input.min_int()
            min_n_samples = input.min_n_spots()
            horizon_size = input.horizon_size()
            additional_params = 'normalization:tic;' if input.normalization() else ''

            if dynamic:
                dyn_res = input.MSI_res()
                dyn_max_retry = input.max_extra_rows()
            else:
                dyn_res = 200
                dyn_max_retry = 5

            # call the function to calculate the core profile
            df_1d = get_msi_depth_profile_from_gui(
                exported_txt_path, sqlite_db_path, target_cmpds, how, spot_method, dynamic, dyn_res, dyn_max_retry, tol,
                min_snr, min_int,
                min_n_samples, horizon_size, save_2d_path, save_1d_path, additional_params, show_tk_message=False,
                return_df_1d=True
            )

            # plot the result
            downcore_fig.set(self.plot_1d(df_1d))

        @output
        @render.plot
        def downcore_profile_fig():
            if downcore_fig() is not None:
                return downcore_fig()
            else:
                return None

        @reactive.Effect
        @reactive.event(input.target_cmpds)
        def _():
            try:
                target_cmpds = input.target_cmpds()
                if target_cmpds:
                    cmpd_names = [cmpd.split(':')[0] for cmpd in target_cmpds.split(';')]
                    if len(cmpd_names) == 1:
                        how_expr = f"data['int_{cmpd_names[0]}'].mean()"
                        session.send_input_message("how", {"value": how_expr})
                    if len(cmpd_names) == 2:
                        how_expr = f"data['int_{cmpd_names[0]}'].sum() / (data['int_{cmpd_names[0]}'].sum() + data['int_{cmpd_names[1]}'].sum())"
                        session.send_input_message("how", {"value": how_expr})
            except Exception as e:
                pass

        @reactive.Effect
        @reactive.event(input.metadata)
        def _():
            try:
                sqlite_db_path = input.metadata()
                base_path = os.path.dirname(sqlite_db_path)
                if os.path.exists(base_path):
                    session.send_input_message("save2d", {"value": os.path.join(base_path, "2D.csv")})
                    session.send_input_message("save1d", {"value": os.path.join(base_path, "1D.csv")})
            except Exception as e:
                pass


def run_app():
    app = MSIDownCoreProfileApp()

    # Run the app in a separate thread to avoid blocking
    def start_server():
        app.app.run(port=61234)

    threading.Thread(target=start_server, daemon=True).start()

    # Give the server a moment to start
    time.sleep(1)

    # Open the app in a web browser
    webbrowser.open("http://localhost:61234")

if __name__ == "__main__":
    pass


import os
import sys
import tkinter as tk
from tkinter import simpledialog, messagebox

import PIL
import numpy as np

from PIL import Image, ImageTk

Image.MAX_IMAGE_PIXELS = None


def get_right_click_event():
    return "<Button-2>" if sys.platform == "darwin" else "<Button-3>"


def key_to_str(key):
    return f"{key[0]},{key[1]}"


def str_to_key(s):
    x_str, y_str = s.split(",")
    return float(x_str), float(y_str)


def draw_teaching_points(x, y, app, size=3, img_tag=None):
    tp_tag = f"tp_{int(x)}_{int(y)}"

    app.canvas.create_oval(
        x - size,
        y - size,
        x + size,
        y + size,
        fill="red",
        tags=tp_tag
    )
    if img_tag is not None:
        app.canvas.addtag_withtag(img_tag, tp_tag)

    app.canvas.tag_bind(tp_tag,
                        get_right_click_event(),
                        lambda e: app.right_click_on_tp.show_menu(e, tp_tag))

    return tp_tag


class LoadedImage:
    """A superclass for the loaded images"""

    def __init__(self):
        self.img_path = None
        self.img = None
        self.thumbnail = None
        self.thumbnail_size = (500, 500)
        self.tk_img = None
        self._tag = None
        self.locked = False
        self.origin = (None, None)

    @property
    def x(self):
        return self.origin[0]

    @property
    def y(self):
        return self.origin[1]

    @property
    def tag(self):
        return self._tag

    def __str__(self):
        return self.img_path

    def __repr__(self):
        return self.img_path

    def load_image(self):
        try:
            self.img = Image.open(self.img_path)
        except PIL.UnidentifiedImageError:
            data = np.loadtxt(self.img_path, delimiter=";")
            data_norm = (255 * (data - np.min(data)) / (np.max(data) - np.min(data))).astype(np.uint8)
            self.img = Image.fromarray(data_norm)
        self.update_thumbnail()

    def update_thumbnail(self):
        self.thumbnail = self.img.copy()
        self.thumbnail.thumbnail(self.thumbnail_size)
        self.tk_img = ImageTk.PhotoImage(self.thumbnail)

    @classmethod
    def from_path(cls, img_path):
        self = cls()
        self.img_path = img_path
        self._tag = os.path.basename(img_path).replace(" ", "_")
        self.load_image()
        return self

    def create_im_on_canvas(self, app):
        if self.origin == (None, None):
            x = app.canvas.canvasx(app.canvas.winfo_width() / 2)
            y = app.canvas.canvasy(app.canvas.winfo_height() / 2)
            self.origin = (x - self.thumbnail.width / 2, y - self.thumbnail.height / 2)

        app.canvas.create_image(
            self.origin[0],
            self.origin[1],
            anchor="nw",
            image=self.tk_img,
            tags=self.tag
        )
        app.bind_events_to_loaded_images(self)

    def resize(self, size):
        self.thumbnail_size = size
        self.update_thumbnail()

    def enlarge(self, scale_factor):
        self.thumbnail_size = (
            self.thumbnail.width * scale_factor,
            self.thumbnail.height * scale_factor
        )
        self.update_thumbnail()

    def lock(self):
        self.locked = True

    def unlock(self):
        self.locked = False

    def to_json(self):
        return {
            "type": self.__class__.__name__,
            "img_path": self.img_path,
            "thumbnail_size": self.thumbnail_size,
            "origin": self.origin,
            "locked": self.locked,
        }

    @classmethod
    def from_json(cls, json_data, app):
        self = cls()
        self.img_path = json_data['img_path']
        self.thumbnail_size = tuple(json_data['thumbnail_size'])
        self.origin = tuple(json_data['origin'])
        self.locked = json_data['locked']
        self._tag = os.path.basename(self.img_path).replace(" ", "_")
        self.load_image()
        self.create_im_on_canvas(app)
        return self

    def rm(self, app):
        if messagebox.askyesno("Remove Image", "Are you sure you want to remove this image?"):
            app.canvas.delete(self.tag)
            del app.items[self.tag]


class TeachableImage(LoadedImage):
    """A subclass of LoadedImage that holds the teachable images"""

    def __init__(self):
        super().__init__()
        self.teaching_points_px_coords = {}
        self.teaching_points = {}
        self.tp_size = 3
        self.show_sediment_start_line_not_set_warning = True

    @property
    def label_indexed_teaching_points(self):
        return {v[3]: v[:3] for v in self.teaching_points.values() if len(v) >= 4}

    @property
    def px_teaching_points_labels(self):
        return {k: v[3] for k, v in self.teaching_points.items() if len(v) >= 4}

    def show_tp_labels(self, app):
        for k, v in self.teaching_points.items():
            if len(v) >= 4:
                label = v[3]
                text = tk.Text(app.canvas, height=1, width=3)
                text.insert(tk.END, label)
                text.config(state=tk.DISABLED)
                app.canvas.create_window(
                    k[0],
                    k[1],
                    window=text,
                    anchor="nw",
                    tags="tp_labels"
                )

    def _update_teaching_points(self, app, transform_func):
        old_keys = list(self.teaching_points.keys())
        for old_key in old_keys:
            x, y = old_key
            app.canvas.delete(f"tp_{int(x)}_{int(y)}")
            x_new, y_new = transform_func(x, y)
            new_key = (x_new, y_new)
            draw_teaching_points(x_new, y_new, app, size=self.tp_size, img_tag=self.tag)
            self.teaching_points[new_key] = self.teaching_points.pop(old_key)
            if self.teaching_points_px_coords:
                self.teaching_points_px_coords[new_key] = self.teaching_points_px_coords.pop(old_key)

    def update_teaching_points_on_resize(self, app, origin, scale_factor):
        self._update_teaching_points(
            app,
            lambda x, y: (
                origin[0] + (x - origin[0]) * scale_factor,
                origin[1] + (y - origin[1]) * scale_factor
            )
        )

    def update_teaching_points(self, app, offset_x, offset_y):
        self._update_teaching_points(
            app,
            lambda x, y: (x + offset_x, y + offset_y)
        )

    def add_teaching_point(self, event, app):
        canvas_x, canvas_y = app.canvas.canvasx(event.x), app.canvas.canvasy(event.y)
        draw_teaching_points(canvas_x, canvas_y, app, size=self.tp_size, img_tag=self.tag)

        if app.sediment_start is not None and app.cm_per_pixel is not None:
            depth = abs(app.canvas.coords(app.sediment_start)[0] - canvas_x) * app.cm_per_pixel
        else:
            if self.show_sediment_start_line_not_set_warning:
                messagebox.showwarning("Warning", "Sediment start line not found, depth not calculated.")
                self.show_sediment_start_line_not_set_warning = False
            depth = None

        original_width, original_height = self.img.size
        scale_x = original_width / self.thumbnail.width
        scale_y = original_height / self.thumbnail.height

        img_x = (canvas_x - self.x) * scale_x
        img_y = (canvas_y - self.y) * scale_y

        teaching_point_data = (img_x, img_y, depth)

        teaching_point_key = (canvas_x, canvas_y)
        self.teaching_points[teaching_point_key] = teaching_point_data
        self.teaching_points_px_coords[teaching_point_key] = teaching_point_data

    def to_json(self):
        json_data = super().to_json()
        json_data["teaching_points"] = {key_to_str(k): v for k, v in self.teaching_points.items()}
        json_data["teaching_points_px_coords"] = {key_to_str(k): v for k, v in self.teaching_points_px_coords.items()}
        return json_data

    @classmethod
    def from_json(cls, json_data, app):
        self = super().from_json(json_data, app)
        self.teaching_points = {str_to_key(k): v for k, v in json_data.get("teaching_points", {}).items()}
        for tp_key in self.teaching_points.keys():
            x, y = tp_key
            draw_teaching_points(x, y, app, size=self.tp_size, img_tag=self.tag)
        self.teaching_points_px_coords = {str_to_key(k): v for k, v in json_data.get("teaching_points_px_coords", {}).items()}
        if len(self.teaching_points_px_coords) == 0:
            self.teaching_points_px_coords = self.teaching_points.copy()
        return self


class MsiImage(TeachableImage):
    """A subclass of TeachableImage that holds the MSI image"""

    def __init__(self):
        super().__init__()
        self.msi_rect = None
        self.px_rect = None
        self.teaching_points_px_coords = {}

    def update_tp_coords(self, sqlite_db_path):
        if not self.teaching_points_px_coords:
            self.teaching_points_px_coords = self.teaching_points.copy()
        if self.msi_rect is None or self.px_rect is None:
            import sqlite3
            conn = sqlite3.connect(sqlite_db_path)
            c = conn.cursor()
            c.execute('SELECT msi_img_file_name, px_rect, msi_rect FROM metadata')
            data = c.fetchall()
            for row in data:
                im_name, px_rect, msi_rect = row
                if im_name == os.path.basename(self.img_path):
                    self.px_rect = eval(px_rect)
                    self.msi_rect = eval(msi_rect)
                    break
            conn.close()
        if self.msi_rect is None or self.px_rect is None:
            messagebox.showerror("Error", "Something went wrong, please check the metadata file")
            return
        if not self.teaching_points:
            messagebox.showerror("Error", "No teaching points found")
            return
        x_min, y_min, x_max, y_max = self.msi_rect
        x_min_px, y_min_px, x_max_px, y_max_px = self.px_rect
        for k, v in self.teaching_points_px_coords.items():
            msi_x = (v[0] - x_min_px) / (x_max_px - x_min_px) * (x_max - x_min) + x_min
            msi_y = (v[1] - y_min_px) / (y_max_px - y_min_px) * (y_max - y_min) + y_min
            self.teaching_points[k] = (msi_x, msi_y, *v[2:])

    def to_json(self):
        json_data = super().to_json()
        json_data["msi_rect"] = self.msi_rect
        json_data["px_rect"] = self.px_rect
        return json_data

    @classmethod
    def from_json(cls, json_data, app):
        self = super().from_json(json_data, app)
        self.msi_rect = tuple(json_data['msi_rect']) if json_data['msi_rect'] is not None else None
        self.px_rect = tuple(json_data['px_rect']) if json_data['px_rect'] is not None else None
        return self


class VerticalLine:
    color_map = {
        "scale_line": "blue",
        "sediment_start_line": "green"
    }

    def __init__(self, position, line_type=None):
        self.position = position
        self.depth = None
        self.line_type = line_type
        self._tag = f"vl_{position[0]}"
        self.type = "VerticalLine"

    @property
    def tag(self):
        return self._tag

    def create_on_canvas(self, app):
        app.canvas.create_line(
            self.position[0],
            0,
            self.position[0],
            5000,
            fill=self.color,
            tags=self.tag,
            width=1,
            dash=(4, 4)
        )
        app.bind_events_to_vertical_lines(self)

    def add_depth_text(self, app, depth):
        self.depth = depth
        text = tk.Text(app.canvas, height=1, width=8)
        text.insert(tk.END, f"{depth}cm")
        text.config(state=tk.DISABLED)
        app.canvas.create_window(
            self.position[0],
            10,
            window=text,
            anchor="nw",
            tags=self.tag
        )

    @property
    def x(self):
        return self.position[0]

    @property
    def y(self):
        return self.position[1]

    @property
    def color(self):
        return self.color_map.get(self.line_type, "red")

    def __str__(self):
        return self.tag

    def __repr__(self):
        return self.tag

    def rm(self, app):
        app.canvas.delete(self.tag)
        del app.items[self.tag]

    @classmethod
    def from_json(cls, json_data, app):
        self = cls(json_data['position'], line_type=json_data.get('line_type'))
        self._tag = json_data['tag']
        self.depth = json_data['depth']
        self.create_on_canvas(app)
        return self

    def to_json(self):
        return {
            "type": "VerticalLine",
            "tag": self.tag,
            "position": self.position,
            "depth": self.depth,
            "line_type": self.line_type
        }


if __name__ == "__main__":
    pass